#!/user/bin/env python
'''rWork.py:

This filter returns True if the r_work value for this structure is within the
specified range

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''


class RWork(object):
    '''This filter returns True if the rWork value for this structure is within
    the specified range.

    Attributes
    ----------
        min_Rwork (float): The lower bound r_work value
        max_Rwork (float): The upper bound r_work value
    '''

    def __init__(self, minRwork, maxRwork):
        self.min_Rwork = minRwork
        self.max_Rwork = maxRwork

    def __call__(self, t):
        if t[1].r_work == None:
            return False

        return t[1].r_work >= self.min_Rwork and t[1].r_work <= self.max_Rwork



#!/usr/bin/env python
'''mmtfReader.py: Methods for reading and downloading structures in MMTF file
formats. The data are returned as a PythonRDD with the structure id (e.g. PDB ID)
as the key and the structural data as the value.

Supported operations and file formats:
    - Read directory of MMTF-Hadoop sequence files in full and reduced representation
    - Download MMTF full and reduced representations using web service (mmtf.rcsb.org)
    - Read directory of MMTF files (.mmtf, mmtf.gz)

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import os
import msgpack
import gzip
from mmtfPyspark.utils import MmtfStructure
from mmtf.api import default_api
from os import path, walk

text = "org.apache.hadoop.io.Text"
byteWritable = "org.apache.hadoop.io.BytesWritable"


def read_full_sequence_file(sc, pdbId=None, fraction=None, seed=123):
    '''Reads a MMTF-Hadoop Sequence file using the default file location.
    The default file location is determined by {mmtfReader.get_mmtf_full_path}

    Downloads
    ---------
        To download mmtf files: {https://mmtf.rcsb.org/download.htm}

    Attributes
    ----------
        sc (Spark Context):
        pdbID (list(str)): List of structures to read
        fraction (float): fraction of structure to read
        seed (int): random seed
    '''
    return read_sequence_file(get_mmtf_full_path(), sc, pdbId, fraction, seed)


def read_reduced_sequence_file(sc, pdbId=None, fraction=None, seed=123):
    '''Reads a MMTF-Hadoop Sequence file using the default file location.
    The default file location is determined by {mmtfReader.get_mmtf_reduced_path}

    Downloads
    ---------
        To download mmtf files: {https://mmtf.rcsb.org/download.htm}

    Attributes
    ----------
        sc (Spark Context):
        pdbID (list(str)): List of structures to read
        fraction (float): fraction of structure to read
        seed (int): random seed
    '''
    return read_sequence_file(get_mmtf_reduced_path(), sc, pdbId, fraction, seed)


def read_sequence_file(path, sc, pdbId=None, fraction=None, seed=123):
    '''Reads an MMTF Hadoop Sequence File. Can read all files from path,
    randomly rample a fraction, or a subset based on input list.
    See <a href="http://mmtf.rcsb.org/download.html"> for file download information</a>

    Attributes
    ----------
        path (str): path to file directory
        sc (Spark Context):
        pdbID (list(str)): List of structures to read
        fraction (float): fraction of structure to read
        seed (int): random seed
    '''

    if not os.path.exists(path):
        raise Exception("file path does not exist")

    infiles = sc.sequenceFile(path, text, byteWritable)

    # Read in all structures from a directory
    if (pdbId == None and fraction == None):
        return infiles.map(_call_sequence_file)

    # Read in a specified list of pdbIds
    elif(pdbId != None and fraction == None):
        pdbIdSet = set(pdbId)
        return infiles.filter(lambda t: str(t[0]) in pdbIdSet).map(_call_sequence_file)

    # Read in a random fraction of structures from a directory
    elif (pdbId == None and fraction != None):
        return infiles.sample(False, fraction, seed).map(_call_sequence_file)

    else:
        raise Exception("Inappropriate combination of parameters")


def read_mmtf_files(path, sc):
    '''Read the specified PDB entries from a MMTF file

    Attributes
    ----------
        path (str): Path to MMTF files
        sc (Spark context)

    Returns
    -------
        structure data as keywork/value pairs
    '''

    if not os.path.exists(path):
        raise Exception("file path does not exist")

    return sc.parallelize(_get_files(path)).map(_call_mmtf).filter(lambda t: t != None)


def download_mmtf_files(pdbIds, sc, reduced=False):
    '''Download and reads the specified PDB entries using
    <a href="http://mmtf.rcsb.org/download.html">MMTF web services</a>.
    with either full or reduced format

    Attributes
    ----------
        path (str): Path to PDB files
        sc (Spark context)
        reduced (bool): flag to indicate reduced or full file format

    Returns
    -------
        structure data as keywork/value pairs
    '''

    return sc.parallelize(set(pdbIds)).map(lambda t: _get_structure(t, reduced))


def download_full_mmtf_files(pdbIds, sc):
    '''Download and reads the specified PDB entries in full mmtf format using
    <a href="http://mmtf.rcsb.org/download.html">MMTF web services</a>.

    Attributes
    ----------
        path (str): Path to PDB files
        sc (Spark context)

    Returns
    -------
        structure data as keywork/value pairs
    '''

    return sc.parallelize(set(pdbIds)).map(lambda t: _get_structure(t, False))


def download_reduced_mmtf_files(pdbIds, sc):
    '''Download and reads the specified PDB entries in reduced mmtf format using
    <a href="http://mmtf.rcsb.org/download.html">MMTF web services</a>.

    Attributes
    ----------
        path (str): Path to PDB files
        sc (Spark context)

    Returns
    -------
        structure data as keywork/value pairs
    '''

    return sc.parallelize(set(pdbIds)).map(lambda t: _get_structure(t, True))


def _get_structure(pdbId, reduced):
    '''Download and decode a list of structure from a list of PDBid

    Attributes
    ----------
        pdbID (List(str)): List of structures to download

    Returns
    -------
        tuble of pdbID and deccoder
    '''

    unpack = default_api.get_raw_data_from_url(pdbId, reduced)
    decoder = MmtfStructure(unpack)
    return (pdbId, decoder)


def _call_sequence_file(t):
    '''Call function for hadoop sequence files'''
    # TODO: check if all sequence files are gzipped
    data = default_api.ungzip_data(t[1])
    unpack = msgpack.unpackb(data.read(), raw=False)
    decoder = MmtfStructure(unpack)
    return (str(t[0]), decoder)


def _call_mmtf(f):
    '''Call function for mmtf files'''

    if ".mmtf.gz" in f:
        name = f.split('/')[-1].split('.')[0].upper()
        data = gzip.open(f, 'rb')
        unpack = msgpack.unpack(data, raw=False)
        decoder = MmtfStructure(unpack)
        return (name, decoder)

    elif ".mmtf" in f:
        name = f.split('/')[-1].split('.')[0].upper()
        unpack = msgpack.unpack(open(f, "rb"), raw=False)
        decoder = MmtfStructure(unpack)
        return (name, decoder)


def _get_files(user_path):
    '''Get List of files from path

    Attributes
    ----------
        user_path (str): File path

    Returns
    -------
        files (List(str)): list of files in path
    '''
    files = []
    for dirpath, dirnames, filenames in walk(user_path):
        for f in filenames:
            if path.isdir(f):
                files += getFiles(f)
            else:
                files.append(dirpath + '/' + f)
    return files


def get_mmtf_full_path():
    '''Returns the path to the full MMTF-Hadoop sequence file.
    It looks for the environmental variable "MMTF_FULL", if not set, an error
    message will be shown.

    Returns
    -------
        path to the mmtf_full directory
    '''

    if 'MMTF_FULL' in os.environ:
        print(
            f"Hadoop Sequence file path: MMTF_FULL={os.environ.get('MMTF_FULL')}")
        return os.environ.get("MMTF_FULL")
    else:
        raise EnvironmentError("Environmental variable 'MMTF_FULL not set'")


def get_mmtf_reduced_path():
    '''Returns the path to the reduced MMTF-Hadoop sequence file.
    It looks for the environmental variable "MMTF_REDUCED", if not set, an error
    message will be shown.

    Returns
    -------
        path to the mmtf_reduced directory
    '''

    if 'MMTF_REDUCED' in os.environ:
        print(
            f"Hadoop Sequence file path: MMTF_REDUCED={os.environ.get('MMTF_REDUCED')}")
        return os.environ.get("MMTF_REDUCED")
    else:
        raise EnvironmentError("Environmental variable 'MMTF_REDUCED not set'")



#!/user/bin/env python
'''containsRnaChain.py

This filter passes entries that contain RNA chains. The default constructor
passes entries that contain at least one RNA chain. If the "exclusive" flag is
set to true in the constructor, all polymer chains must be RNA. For a multi-model
structure (e.g., NMR structure), this filter only checks the first model.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''
from mmtfPyspark.filters import ContainsPolymerChainType


class ContainsRnaChain(object):
    '''Default constructor matches any entry that contains at least one RNA
    chain. As an example, an RNA-protein complex passes this filter.

    Optional constructor that can be used to filter entries that exclusively
    contain DNA chains. For example, with "exclusive" set to true, an
    RNA-protein complex complex does not pass this filter.

    Attributes
    ----------
        exclusive (bool) if true, only return entries that contain RNA chains
    '''

    def __init__(self, exclusive=False):
        self.filter = ContainsPolymerChainType(
            ContainsPolymerChainType.RNA_LINKING, exclusive)

    def __call__(self, t):
        return self.filter(t)



from .proteinSequenceEncoder import ProteinSequenceEncoder
from .sparkMultiClassClassifier import SparkMultiClassClassifier
from .sparkRegressor import SparkRegressor
from .sparkMultiClassClassifier import SparkMultiClassClassifier
from . import datasetBalancer, pythonRDDToDataset, sequenceNgrammer



#!/user/bin/env python
'''polymerInteractionFIngerprint.py

Finds interactions between polymer chains and maps them onto polymer sequences.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

from mmtfPyspark.interactions import InteractionFilter, AtomInteraction, InteractionCenter
from mmtfPyspark.utils import ColumnarStructure
from mmtfPyspark.utils import DistanceBox
from pyspark.sql import Row
import numpy as np


class PolymerInteractionFingerprint(object):

    def __init__(self, interactionFilter):
        self.filter = interactionFilter

    def __call__(self, t):

        structureId = t[0]
        structure = t[1]
        return self.get_interactions(structureId, structure)

    def get_interactions(self, structureId, structure):
        rows = []

        cutoffDistanceSquared = self.filter.get_distance_cutoff() ** 2
        arrays = ColumnarStructure(structure, True)

        chainNames = arrays.get_chain_names()
        groupNames = arrays.get_group_names()
        groupNumbers = arrays.get_group_numbers()
        atomNames = arrays.get_atom_names()
        entityIndices = arrays.get_entity_indices()
        elements = arrays.get_elements()
        polymer = arrays.is_polymer()

        sequenceMapIndices = arrays.get_sequence_positions()
        x = arrays.get_x_coords()
        y = arrays.get_y_coords()
        z = arrays.get_z_coords()

        # create a distance box for quick lookup interactions of polymer atoms
        # of the specified elements
        boxes = {}
        for i in range(arrays.get_num_atoms()):

            if polymer[i] \
                and (self.filter.is_target_group(groupNames[i]) or self.filter.is_query_group(groupNames[i])) \
                and (self.filter.is_target_atom_name(atomNames[i]) or self.filter.is_query_atom_name(atomNames[i])) \
                and (self.filter.is_target_element(elements[i]) or self.filter_is_query_element_name(elements[i])) \
                and not self.filter.is_prohibited_target_group(groupNames[i]):

                if chainNames[i] not in boxes:
                    box = DistanceBox(self.filter.get_distance_cutoff())
                    boxes[chainNames[i]] = box

                newPoint = np.array([x[i],y[i],z[i]])
                boxes[chainNames[i]].add_point(newPoint,i)

        chainBoxes = [(k,v) for k,v in boxes.items()]

        # loop over all pairwise polymer chain interactions
        for i in range(len(chainBoxes) - 1):
            chainI = chainBoxes[i][0]
            boxI = chainBoxes[i][1]

            for j in range(i+1, len(chainBoxes)):
                chainJ = chainBoxes[j][0]
                boxJ = chainBoxes[j][1]

                intersectionI = boxI.getIntersection(boxJ)
                intersectionJ = boxJ.getIntersection(boxI)

                # maps to store sequence indices mapped to group numbers
                indicesI = {}
                indicesJ = {}

                entityIndexI = -1
                entityIndexJ = -1

                # loop over pairs of atom interactions and check if
                # they satisfy the interaction filter criteria

                for n in intersectionI:

                    for m in intersectionJ:

                        dx = x[n] - x[m]
                        dy = y[n] - y[m]
                        dz = z[n] - z[m]
                        dSq = dx * dx + dy * dy + dz * dz

                        if dSq <= cutoffDistanceSquared:
                            if self.filter.is_target_group(groupNames[n]) \
                                and self.filter.is_target_atom_name(atomNames[n]) \
                                and self.filter.is_target_element(elements[n]) \
                                and self.filter.is_query_group(groupNames[m]) \
                                and self.filter.is_query_atom_name(atomNames[m]) \
                                and self.filter.is_query_element(elements[m]):

                                entityIndexI = entityIndices[n]
                                indicesI[sequenceMapIndices[n]] = groupNumbers[n]

                            if self.filter.is_target_group(groupNames[m]) \
                                and self.filter.is_target_atom_name(atomNames[m]) \
                                and self.filter.is_target_element(elements[m]) \
                                and self.filter.is_query_group(groupNames[n]) \
                                and self.filter.is_query_atom_name(atomNames[n]) \
                                and self.filter.is_query_element(elements[n]):

                                entityIndexJ = entityIndices[m]
                                indicesJ[sequenceMapIndices[m]] = groupNumbers[m]

            if len(indicesI) >= self.filter.get_min_interactions():
                sequenceIndiciesI = sorted([int(i) for i in indicesI.keys()])
                groupNumbersI = sorted(list(indicesI.values()))

                rows.append(Row(structureId + '.' + chainI, chainJ, chainI, \
                                groupNumbersI, sequenceIndiciesI, \
                                structure.entity_list[entityIndexI]['sequence']))

            if len(indicesJ) >= self.filter.get_min_interactions():
                sequenceIndiciesJ = sorted([int(i) for i in indicesJ.keys()])
                groupNumbersJ = sorted(list(indicesJ.values()))

                rows.append(Row(structureId + '.' + chainJ, chainI, chainJ, \
                                groupNumbersJ, sequenceIndiciesJ, \
                                structure.entity_list[entityIndexJ]['sequence']))

        return rows



#!/user/bin/env python
'''releaseDate.py

this filter returns true if the release date for this
structure is within the specified range

authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

from dateutil.parser import parse


class ReleaseDate(object):

    def __init__(self, startDate, endDate):
        '''This filter retuns true if the release date for the structure is
        within the specified range.

        Attributes
        ----------
            startDate (str): start of the release date range
            enddate (str): end of the the release date range
        '''
        self.startDate = parse(startDate)
        self.endDate = parse(endDate)

    def __call__(self, t):
        structure = t[1]
        releaseDate = parse(structure.release_date)

        return releaseDate >= self.startDate and releaseDate <= self.endDate



#!/usr/bin/env python
'''traverseStructureHierachy.py

A class that prints of hierachy information about a structure

Example
-------

    pdb = mmtfReader.download_mmtf_files(['1STP'], sc)
    pdb.foreach(lambda t: traverseStructureHierarchy.printMmtfInfo(t[1]) )

    structure = mmtfReader.download_mmtf_files(['1STP'], sc).collect()[0]
    traverseStructureHierarchy.print_mmtf_info(structure[1])

Authorship information:
    __author__ = "Yue Yu"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from mmtfPyspark.utils.dsspSecondaryStructure import *

def _get_chain_to_entity_index(structure):
    entityChainIndex = [0] * structure.num_chains
    for i in range(0, len(structure.entity_list)):
        for j in structure.entity_list[i]["chainIndexList"]:
            entityChainIndex[j] = i
    return entityChainIndex


def _list_to_string(temp):
    return("[" + ", ".join(map(str, temp)) + "]")

def _check_structure_or_tuple(structure):
    if type(structure) == tuple:
        structure = structure[1]
    return structure


def print_mmtf_info(structure):
    structure = _check_structure_or_tuple(structure)

    print("*** STRUCTURE INFO ***")
    print("MmtfProducer    : " + structure.mmtf_producer)
    print("MmtfVersion     : " + structure.mmtf_version)
    print();

def print_metadata(structure):
    structure = _check_structure_or_tuple(structure)

    print("*** METADATA ***")
    print("StructureId           : " + structure.structure_id)
    print("Title                 : " + structure.title)
    print("Deposition date       : " + structure.deposition_date)
    print("Release date          : " + structure.release_date)
    print("Experimental method(s): " + _list_to_string(structure.experimental_methods))
    print("Resolution            : " + str(structure.resolution))
    print("Rfree                 : " + str(structure.r_free))
    print("Rwork                 : " + str(structure.r_work))
    print()


def print_crystallographic_data(structure):
    structure = _check_structure_or_tuple(structure)

    print("*** CRYSTALLOGRAPHIC DATA ***")
    print("Space group           : " + structure.space_group)
    print("Unit cell dimensions  : " + _list_to_string(["%0.2f" % i for i in structure.unit_cell]))
    print()

def print_bioassembly_data(structure):
    structure = _check_structure_or_tuple(structure)

    print("*** BIOASSEMBLY DATA ***")
    print("Number bioassemblies: " + str(len(structure.bio_assembly)))
    for i in range(0, len(structure.bio_assembly)):
        print("bioassembly: " + structure.bio_assembly[i]["name"])
        transformations = structure.bio_assembly[i]["transformList"]
        print("  Number transformations: " + str(len(transformations)))
        for j in range(0, len(transformations)):
            print("    transformation: " + str(j))
            print("    chains:         " + str(transformations[j]["chainIndexList"]))
            print("    rotTransMatrix: " + str(transformations[j]["matrix"]))

def print_entity_info(structure):
    '''Prints information about unique entities (molecules) in a structure
    '''
    print("*** ENTITY DATA ***")

    for i,e in enumerate(structure.entity_list):
        print(f"entity type            : {i} {e['type']}")
        print(f"entity description     : {i} {e['description']}")
        print(f"entity sequence        : {i} {e['sequence']}")
    print('\n')


def print_structure_data(structure):
    structure = _check_structure_or_tuple(structure)

    print("*** STRUCTURE DATA ***")
    print("Number of models : " + str(structure.num_models))
    print("Number of chains : " + str(structure.num_chains))
    print("Number of groups : " + str(structure.num_groups))
    print("Number of atoms : " + str(structure.num_atoms))
    print("Number of bonds : " + str(structure.num_bonds))
    print()

def print_chain_info(structure):
    structure = _check_structure_or_tuple(structure)

    print("*** CHAIN DATA ***")
    print("Number of chains: " + str(structure.num_chains))
    chainIndex = 0
    for i in range(0, structure.num_models):
        print("model: " + str(i+1))
        for j in range(0, structure.chains_per_model[i]):
            chainName = structure.chain_name_list[chainIndex]
            chainId = structure.chain_id_list[chainIndex]
            groups = structure.groups_per_chain[chainIndex]
            print("chainName: " + chainName + ", chainId: " + chainId + ", groups: " + str(groups))
            chainIndex = chainIndex + 1
    print()

def print_chain_group_info(structure):
    structure = _check_structure_or_tuple(structure)
    structure = structure.set_alt_loc_list()

    print("*** CHAIN AND GROUP DATA ***")
    chainIndex = 0
    groupIndex = 0
    for i in range(0, structure.num_models):
        print("model: " + str(i+1))
        for j in range(0, structure.chains_per_model[i]):
            chainName = structure.chain_name_list[chainIndex]
            chainId = structure.chain_id_list[chainIndex]
            groups = structure.groups_per_chain[chainIndex]
            print("chainName: " + chainName + ", chainId: " + chainId + ", groups: " + str(groups))
            for k in range(0, structure.groups_per_chain[chainIndex]):
                groupId = structure.group_id_list[groupIndex]
                insertionCode = structure.ins_code_list[groupIndex]
                secStruct = structure.sec_struct_list[groupIndex]
                seqIndex = structure.sequence_index_list[groupIndex]

                groupType = structure.group_type_list[groupIndex]

                groupName = structure.group_list[groupType]["groupName"]
                chemCompType = structure.group_list[groupType]["chemCompType"]
                oneLetterCode = structure.group_list[groupType]["singleLetterCode"]
                numAtoms = len(structure.group_list[groupType]["atomNameList"])
                numBonds = len(structure.group_list[groupType]["bondOrderList"])

                print("   groupName      : " + groupName)
                print("   oneLetterCode  : " + oneLetterCode)
                print("   seq. index     : " + str(seqIndex))
                print("   numAtoms       : " + str(numAtoms))
                print("   numBonds       : " + str(numBonds))
                print("   chemCompType   : " + chemCompType)
                print("   groupId        : " + str(groupId))
                print("   insertionCode  : " + insertionCode)
                print("   DSSP secStruct.: " + DsspSecondaryStructure.get_dssp_code(secStruct).get_one_letter_code())

                print()
                groupIndex = groupIndex + 1
            chainIndex = chainIndex + 1
    print()

def print_chain_entity_group_atom_info(structure):
    structure = _check_structure_or_tuple(structure)

    print("*** CHAIN ENTITY GROUP ATOM DATA ***")
    chainToEntityIndex = _get_chain_to_entity_index(structure)
    chainIndex = 0
    groupIndex = 0
    atomIndex = 0
    for i in range(0, structure.num_models):
        print("model: " + str(i+1))
        for j in range(0, structure.chains_per_model[i]):
            chainName = structure.chain_name_list[chainIndex]
            chainId = structure.chain_id_list[chainIndex]
            groups = structure.groups_per_chain[chainIndex]
            print("chainName: " + chainName + ", chainId: " + chainId + ", groups: " + str(groups))

            entityType = structure.entity_list[chainToEntityIndex[chainIndex]]["type"]
            entityDescription = structure.entity_list[chainToEntityIndex[chainIndex]]["description"]
            entitySequence = structure.entity_list[chainToEntityIndex[chainIndex]]["sequence"]
            print("entity type          : " + entityType);
            print("entity description   : " + entityDescription);
            print("entity sequence      : " + entitySequence);

            for k in range(0, structure.groups_per_chain[chainIndex]):
                groupId = structure.group_id_list[groupIndex]
                insertionCode = structure.ins_code_list[groupIndex]
                secStruct = structure.sec_struct_list[groupIndex]
                seqIndex = structure.sequence_index_list[groupIndex]

                groupType = structure.group_type_list[groupIndex]

                groupName = structure.group_list[groupType]["groupName"]
                chemCompType = structure.group_list[groupType]["chemCompType"]
                oneLetterCode = structure.group_list[groupType]["singleLetterCode"]
                numAtoms = len(structure.group_list[groupType]["atomNameList"])
                numBonds = len(structure.group_list[groupType]["bondOrderList"])

                print("   groupName      : " + groupName)
                print("   oneLetterCode  : " + oneLetterCode)
                print("   seq. index     : " + str(seqIndex))
                print("   numAtoms       : " + str(numAtoms))
                print("   numBonds       : " + str(numBonds))
                print("   chemCompType   : " + chemCompType)
                print("   groupId        : " + str(groupId))
                print("   insertionCode  : " + insertionCode)
                print("   DSSP secStruct.: " + DsspSecondaryStructure.get_dssp_code(secStruct).get_one_letter_code())
                print("   Atoms          : ")

                for m in range(0, (len(structure.group_list[groupType]["atomNameList"]))):
                    atomId = structure.atom_id_list[atomIndex]

                    if not structure.alt_loc_set:
                        structure = structure.set_alt_loc_list()

                    altLocId = structure.alt_loc_list[atomIndex]
                    x = structure.x_coord_list[atomIndex]
                    y = structure.y_coord_list[atomIndex]
                    z = structure.z_coord_list[atomIndex]
                    occupancy = structure.occupancy_list[atomIndex]
                    bFactor = structure.b_factor_list[atomIndex]

                    atomName = structure.group_list[groupType]["atomNameList"][m]
                    element = structure.group_list[groupType]["elementList"][m]

                    print("      " + str(atomId) + "\t" + atomName + "\t" + str(altLocId) +
                        "\t" + str(x) + "\t" + str(y) + "\t" + str(z) +
                        "\t" + str(occupancy) + "\t" + str(bFactor) + "\t" + str(element))
                    atomIndex = atomIndex + 1


                groupIndex = groupIndex + 1
            chainIndex = chainIndex + 1
    print('\n')

def print_all_data(structure):
    structure = _check_structure_or_tuple(structure)
    print_mmtf_info(structure)
    print_metadata(structure)
    print_crystallographic_data(structure)
    print_bioassembly_data(structure)
    print_structure_data(structure)
    print_chain_info(structure)
    print_chain_group_info(structure)
    print_chain_entity_group_atom_info(structure)



#!/user/bin/env python
'''structureToAtomInteraction.py

Finds interactions that match the criteria specified by the InteractionFilter

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

from mmtfPyspark.interactions import InteractionFilter, AtomInteraction, InteractionCenter
from mmtfPyspark.utils import ColumnarStructureX
from mmtfPyspark.utils import DistanceBox
import numpy as np


class StructureToAtomInteractions(object):
    '''Class that finds structure to atom intteractions.

    Attributes:
        bfilter (Class): Specifies the conditions for calculating interactions
        pairwise (bool): If True, results as one row per pair interactions.
                         If False, the interactions of one atom with all other
                         atoms are returned as a single row.
    '''

    def __init__(self, bfilter, pairwise=False):

        self.filter = bfilter.value
        self.pairwise = pairwise

    def __call__(self, t):

        interactions = []
        structureId = t[0]

        # convert structure to an array-based format for efficient processing
        arrays = ColumnarStructureX(t[1], True)

        # create a list of query atoms for which interactions should be calculated
        queryAtomIndices = self._get_query_atom_indices(arrays)

        if len(queryAtomIndices) == 0:
            return interactions

        # Add atom (indices) on grid for rapid indexing of atom neighbors on a
        # grid based on a cutoff distance
        box = self._get_distance_box(arrays)

        for queryAtomIndex in queryAtomIndices:
            # find interactions of query atom specified by atom index
            interaction = self._get_interactions(arrays, queryAtomIndex, box)
            interaction.set_structure_id(structureId)

            # only add interations that are within the given limits of interations
            if interaction.get_num_interactions() >= self.filter.get_min_interactions() \
                    and interaction.get_num_interactions() <= self.filter.get_max_interactions():

                # return interactions as either pairs or all interaction of
                # one atom as a row
                if self.pairwise:
                    interactions += interaction.get_pair_interactions_as_rows()
                else:
                    multiInteract = interaction.get_multiple_interactions_as_row(
                        self.filter.get_max_interactions())
                    interactions += multiInteract

        return interactions

    def _get_interactions(self, arrays, queryAtomIndex, box):
        '''Get the interacting neighbors of an atom in a structure

        Attributes
        ----------
            arrays (columnarStructure): structure in columnarStructure format
            queryAtomIndex (int): the index of the querying atom
            box (distanceBox): the distance box of the query atom

        Returns
        -------
            an AtomInteraction class with interacting neighbors
        '''
        interaction = AtomInteraction()

        # get the x,y,z coordinates of the structure
        x = arrays.get_x_coords()
        y = arrays.get_y_coords()
        z = arrays.get_z_coords()

        # get the query atom coordinates
        qx = x[queryAtomIndex]
        qy = y[queryAtomIndex]
        qz = z[queryAtomIndex]

        # get required information of the columnarStructure
        atomToGroupIndices = arrays.get_atom_to_group_indices()
        occupancies = arrays.get_occupancies()
        normalizedbFactors = arrays.get_normalized_b_factors()
        groupNames = arrays.get_group_names()

        # record query atom info
        queryCenter = InteractionCenter(arrays, queryAtomIndex)
        interaction.set_center(queryCenter)

        # Retrieve atom indices of atoms that lay within grid cubes that are
        # within cutoff distance of the query atom
        cutoffDistanceSq = self.filter.get_distance_cutoff() ** 2

        # Retrieve atom indices of atoms that lay within grid cubes
        # that are within cutoff distance of the query atom
        neighborIndices = box.get_neighbors(np.array([qx, qy, qz]))
        # TEST: flattern neighborIndices
        if type(neighborIndices[0]) == list:
            neighborIndices = [
                n for neighbors in neighborIndices for n in neighbors]

        # determine and record interactions with neighbor atoms
        for neighborIndex in neighborIndices:

            # exclude self interactions with a group
            if atomToGroupIndices[neighborIndex] == atomToGroupIndices[queryAtomIndex]:
                continue

            # check if interaction is within distance cutoff
            dx = qx - x[neighborIndex]
            dy = qy - y[neighborIndex]
            dz = qz - z[neighborIndex]
            distSq = dx * dx + dy * dy + dz * dz

            if distSq <= cutoffDistanceSq:
                # Exclude interactions with undesired groups and
                # atoms with partial occupancy (< 1.0)
                if self.filter.is_prohibited_target_group(groupNames[neighborIndex]) \
                        or self.filter.get_normalized_b_factor_cutoff() < normalizedbFactors[neighborIndex] \
                        or occupancies[neighborIndex] < float(1.0):

                    # return an empty atom interaction
                    return AtomInteraction()

                # add interacting atom info
                neighbor = InteractionCenter(arrays, neighborIndex)
                interaction.add_neighbor(neighbor)

                # terminate early if the number of interactions exceeds limit
                if interaction.get_num_interactions() > self.filter.get_max_interactions():
                    return interaction

        return interaction

    def _get_distance_box(self, arrays):
        '''Add atom indices on grid for rapid indexing of atom neighbors on a
        grid based on a cutoff distance

        Attributes
        ----------
            arrays (columnarStructure): structure in columnarStructure format
        '''

        # Get required data
        x = arrays.get_x_coords()
        y = arrays.get_y_coords()
        z = arrays.get_z_coords()
        elements = arrays.get_elements()
        groupNames = arrays.get_group_names()

        box = DistanceBox(self.filter.get_distance_cutoff())
        for i in range(arrays.get_num_atoms()):
            if self.filter.is_target_group(groupNames[i]) \
                    and self.filter.is_target_element(elements[i]):

                newPoint = np.array([x[i], y[i], z[i]])
                box.add_point(newPoint, i)

        return box

    def _get_query_atom_indices(self, arrays):
        '''Returns a list of indices to query atoms in the structure

        Attributes
        ----------
            arrays (columnarStructure): structure in columnarStructure format
        '''

        # Get required data
        groupNames = arrays.get_group_names()
        elements = arrays.get_elements()
        groupStartIndices = arrays.get_group_to_atom_indices()
        occupancies = arrays.get_occupancies()
        normalizedbFactors = arrays.get_normalized_b_factors()

        # Find atoms that match the query criteria and exlcued atoms with
        # partial occupancy
        indices = []
        for i in range(arrays.get_num_groups()):

            start = groupStartIndices[i]
            end = groupStartIndices[i + 1]

            if self.filter.is_query_group(groupNames[start]):
                indices += [j for j in range(start, end)
                            if self.filter.is_query_element(elements[j])
                            and normalizedbFactors[j] < self.filter.get_normalized_b_factor_cutoff()
                            and occupancies[j] >= 1.0]

        return indices



#!/user/bin/env python
'''secondaryStructureElementExtractor.py

Returns a datset of continuous segments of protein sequence with the specified
DSSP secondary structure code (E, H, C) of a minimum length.

Example
-------
    +-------------+-----+
    |sequence     |label|
    +-------------+-----+
    |TFIVTA       |E    |
    |ALTGTYE      |E    |


Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from mmtfPyspark.ml import pythonRDDToDataset
from mmtfPyspark.mappers import StructureToSecondaryStructureElements
from mmtfPyspark.datasets import secondaryStructureExtractor


def get_dataset(structure, label, length=None):
    '''Returns a dataset of continuous segments of protein sequence with the
    specified DSSP secondary structure code (E, H, C) of a minimum length.

    Attributes
    ----------
        structure: structure data
        label (char): DSSP secondary structure label (E, H, C)
        length (int): minimum length of secondary structure segment

    Returns
    -------
        dataset of continuous segments of protein sequence
    '''

    colNames = ["sequence", "label"]

    if length == None:

        rows = secondaryStructureExtractor.get_python_rdd(structure) \
            .flatMap(StructureToSecondaryStructureElements(label))

        return pythonRDDToDataset.get_dataset(rows, colNames)
    else:

        rows = secondaryStructureExtractor.get_python_rdd(structure) \
            .flatMap(StructureToSecondaryStructureElements(label, length))

        return pythonRDDToDataset.get_dataset(rows, colNames)



'''customReportQuery.py

This filter runs an SQL query on specified PDB metadata and annotation fields retrived using
RCSB PDB RESTful web services. The fields are then queried and the resulting PDB IDs are
used to filter the data. The input to the filter consists of an SQL WHERE clause, and list
data columns availible from RCSB PDB web services.

References
----------
    List of supported field names:
        http://www.rcsb.org/pdb/results/reportField.do
    Examples of SQL WHERE clauses:
        https://www.w3schools.com/sql/sql_where.asp


<p>Example: find PDB entries with Enzyme classification number 2.7.11.1
and source organism Homo sapiens:

<pre><code>
    JavaPairRDD<String, StructureDataInterface> pdb = ...
    String whereClause = "WHERE ecNo='2.7.11.1' AND source='Homo sapiens'";
    pdb = pdb.filter(new RcsbWebserviceFilter(whereClause, "ecNo","source"));
</code></pre>

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com:
    __status__ = "Done"
'''

from mmtfPyspark.datasets import customReportService
from pyspark.sql import SparkSession


class CustomReportQuery(object):
    '''Filters using an SQL query on the specified fields

    Attributes
    ----------
        whereClause : WHERE Clause of SQL statement
        fields : one or more field names to be used in query
    '''

    def __init__(self, whereClause, fields):

        # Check if fields are in a list or string
        if type(fields) == str:

            if ',' in fields:
                fields = fields.split(',')

            else:
                fields = [fields]

        # Get requested data columns
        dataset = customReportService.get_dataset(fields)

        # Check if the results contain chain level data
        self.chainLevel = "structureChainId" in dataset.columns

        # Create a temporary view of the dataset
        dataset.createOrReplaceTempView("table")

        # Create SparkSession
        spark = SparkSession.builder.getOrCreate()

        # Run SQL query
        if (self.chainLevel):
            # For chain level data
            sql = "SELECT structureChainID, structureId, chainId FROM table " \
                  + whereClause
            results = spark.sql(sql)

            # Add both PDB entry and chain level data, so chain-based data can be filtered
            self.pdbIds = results.distinct().rdd.map(lambda x: x[0]).collect()
            self.pdbIds += results.distinct().rdd.map(lambda x: x[1]).collect()

        else:
            # For PDB entry level data
            sql = "SELECT structureId FROM table " + whereCaluse
            results = spark.sql(sql)
            self.pdbIds = results.distinct().rdd.map(lambda x: x[0]).collect()

        self.pdbIds = list(set(self.pdbIds))

    def __call__(self, t):

        match = t[0] in self.pdbIds

        # If results are PDB IDs, but the keys contains chain names,
        # Then truncate the chain name before matching (e.g., 4HHB.A -> 4HHB)
        if (not self.chainLevel) and (not match) and (len(t[0]) > 4):
            print(t[0])
            return t[0][:4] in self.pdbIds

        return match



from .groupInteractionExtractor import groupInteractionExtractor
from . import customReportService, drugBankDataset, g2sDataset, jpredDataset, myVariantDataset, pdbjMineDataset, polymerSequenceExtractor, secondaryStructureElementExtractor, secondaryStructureExtractor, secondaryStructureSegmentExtractor, swissModelDataset, uniProt



#!/user/bin/env python
'''orFilter.py

This filter wraps two filter and returns true if one of the filters passes

authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''


class OrFilter(object):
    '''Constructor takes another filter as input

    Attributes
    ----------
        filter1 (filter): first filter to be negated
        filter2 (filter): second filter to be negated
    '''

    def __init__(self, filter1, filter2):
        self.filter1 = filter1
        self.filter2 = filter2

    def __call__(self, t):
        return self.filter1(t) or self.filter2(t)



#!/user/bin/env python
'''groupInteractionExtractor.py

Creates a dataset of noncovalent interactions of specified groups (residues)
in macromolecular structures. The criteria for interactions are specified using
an InteractionFilter. The interactions can be returned as interacting atom pairs
or as one row per interacting atom.

Typical use cases include:
    Find interactions between a metal ion and protein/DNA/RNA
    Find interactions between a small molecule and protein/DNA/RNA

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

from pyspark.sql import SparkSession
from pyspark import SparkContext
from mmtfPyspark.interactions import StructureToAtomInteractions, AtomInteraction


class GroupInteractionExtractor(object):

    def get_pair_interactions(self, structures, interactionFilter):
        '''Returns a Dataset of pairwise interactions that satisfy the criteria of
        the InteractionFilter. Each atom, its interacting neightbor atom, and
        the interacting distance is represented as a row.

        Attributes
        ----------
            structures (PythonRDD): a set of PDB structures
            interactionFilter (InteractionFilter): filter criteria for determing
                                                   noncovalent interactions

        Returns
        -------
            Dataset of pairwise interactions
        '''

        spark = SparkSession.builder.getOrCreate()
        sc = spark.sparkContext

        # calculate interactions
        pairwise = True
        rows = structures.flatMap(StructureToAtomInteractions(
            sc.broadcast(interactionFilter), pairwise))

        # convert PythonRDD to Dataset
        return spark.createDataFrame(rows, AtomInteraction().get_pair_interaction_schema())

    def get_interactions(self, structures, interactionFilter):
        '''Returns a dataset of interactions that satisfy the criteria of the
        InteractionFilter. each atom and its interacting neightbor atoms are
        represented as a row in a Dataset. In addition, geometric freatures of
        the interactions, such as distances, angles, and orientation order
        parameters are returned in each row.

        Attributes:
            structures (PythonRDD): a set of PDB structures
            interactionFilter (InteractionFilter): filter criteria for determing
                                                   noncovalent interactions

        Returns:
            Dataset of pairwise interactions
        '''

        spark = SparkSession.builder.getOrCreate()
        sc = spark.sparkContext

        # calculate interactions
        pairwise = False
        rows = structures.flatMap(StructureToAtomInteractions(
            sc.broadcast(interactionFilter), pairwise))

        # convert PythonRDD to Dataset
        return spark.createDataFrame(rows, AtomInteraction().get_schema(interactionFilter.get_max_interactions()))



#!/user/bin/env python
'''structureToPolymerChain.py:

Maps a structure to its individual polymer chains. Polymer chains
include polypeptides, polynucleotides, and linear and branched polysaccharides.
For a multi-model structure, only the first model is considered.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "debug"
'''
from mmtf.utils import *
from mmtf.api.mmtf_writer import MMTFEncoder
from mmtfPyspark.utils import MmtfStructure

class StructureToPolymerChains(object):
    '''Extracts all polymer chains from a structure. If the argument is set to true,
	the assigned key is: <PDB ID.Chain ID>, where Chain ID is the unique identifier
	assigned to each molecular entity in an mmCIF file. This Chain ID corresponds to
	<a href="http://mmcif.wwpdb.org/dictionaries/mmcif_mdb.dic/Items/_atom_site.label_asym_id.html">
	_atom_site.label_asym_id</a> field in an mmCIF file.

    Attributes
    ----------
        useChainIdInsteadOfChainName (bool): if true, use Chain Id in the key assignments
        excludeDuplicates (bool): if true return only one chain for each sequence
    '''

    def __init__(self, useChainIdInsteadOfChainName = False, excludeDuplicates = False):
        self.useChainIdInsteadOfChainName = useChainIdInsteadOfChainName
        self.excludeDuplicates = excludeDuplicates


    def __call__(self, t):

        if type(t[1]) == MmtfStructure and t[1].alt_loc_set == False:
            structure = t[1].set_alt_loc_list()
        else:
            structure = t[1]

        # Precalculate indices
        numChains = structure.chains_per_model[0]
        chainToEntityIndex = self._get_chain_to_entity_index(structure)
        atomsPerChain, bondsPerChain = self._get_num_atoms_and_bonds(structure)

        chainList = list()
        seqSet = set()
        groupCounter = 0
        atomCounter = 0

        for i in range(numChains):
            polymerChain = MMTFEncoder()

            entityToChainIndex = chainToEntityIndex[i]

            chain_type = structure.entity_list[entityToChainIndex]['type']

            polymer = chain_type == "polymer"
            polymerAtomCount = 0

            atomMap = {}

            structureId = ''

            if polymer:
                # To avoid of information loss, add chainName/IDs and entity id
                # This required by some queries
                structureId = structure.structure_id + '.' +\
                              structure.chain_name_list[i] + '.' +\
                              structure.chain_id_list[i] + '.' +\
                              str(entityToChainIndex + 1)

                # Set header
                polymerChain.init_structure(bondsPerChain[i], atomsPerChain[i],
                    structure.groups_per_chain[i], 1, 1, structureId)
                decoder_utils.add_xtalographic_info(structure, polymerChain)
                decoder_utils.add_header_info(structure, polymerChain)

                # Set model info (only one model: 0)
                polymerChain.set_model_info(0,1)

                # Set entity and chain info
                polymerChain.set_entity_info([0],
                    structure.entity_list[entityToChainIndex]['sequence'],
                    structure.entity_list[entityToChainIndex]['description'],
                    structure.entity_list[entityToChainIndex]['type'])
                polymerChain.set_chain_info(structure.chain_id_list[i],
                    structure.chain_name_list[i],
                    structure.groups_per_chain[i])

            for j in range(structure.groups_per_chain[i]):
                groupIndex = structure.group_type_list[groupCounter]

                if polymer:
                    # Set group info
                    polymerChain.set_group_info(structure.group_list[groupIndex]['groupName'],
                        structure.group_id_list[groupCounter],
                        structure.ins_code_list[groupCounter],
                        structure.group_list[groupIndex]['chemCompType'],
                        len(structure.group_list[groupIndex]['atomNameList']),
                        len(structure.group_list[groupIndex]['bondOrderList']),
                        structure.group_list[groupIndex]['singleLetterCode'],
                        structure.sequence_index_list[groupCounter],
                        structure.sec_struct_list[groupCounter])

                for k in range(len(structure.group_list[groupIndex]['atomNameList'])):
                    if polymer:
                        atomMap[atomCounter] = polymerAtomCount
                        polymerAtomCount += 1

                        polymerChain.set_atom_info(
                            structure.group_list[groupIndex]['atomNameList'][k],
                            structure.atom_id_list[atomCounter],
                            structure.alt_loc_list[atomCounter],
                            structure.x_coord_list[atomCounter],
                            structure.y_coord_list[atomCounter],
                            structure.z_coord_list[atomCounter],
                            structure.occupancy_list[atomCounter],
                            structure.b_factor_list[atomCounter],
                            structure.group_list[groupIndex]['elementList'][k],
                            structure.group_list[groupIndex]['formalChargeList'][k],)

                    atomCounter += 1

                if polymer:
                    # Add intra-group bond info
                    for l in range(len(structure.group_list[groupIndex]['bondOrderList'])):
                        bondIndOne = structure.group_list[groupIndex]['bondAtomList'][l*2]
                        bondIndTwo = structure.group_list[groupIndex]['bondAtomList'][l*2+1]
                        bondOrder = structure.group_list[groupIndex]['bondOrderList'][l]

                        polymerChain.set_group_bond(bondIndOne, bondIndTwo, bondOrder)

                groupCounter += 1

            if polymer:
                # TODO skipping adding inter group bond info for now

                polymerChain.finalize_structure()

                chId = structure.chain_name_list[i]
                if self.useChainIdInsteadOfChainName :
                    chId = structure.chain_id_list[i]
                if self.excludeDuplicates:
                    if chainToEntityIndex[i] in seqSet:
                        continue
                    seqSet.add(chainToEntityIndex[i])
                chainList.append((structure.structure_id + "." + chId, polymerChain))

        return chainList


    def _get_num_atoms_and_bonds(self, structure):
        '''Gets the number of atoms and bonds per chain
        '''
        numChains = structure.chains_per_model[0]
        atomsPerChain = [0] * numChains
        bondsPerChain = [0] * numChains
        groupCounter = 0

        for i in range(numChains):

            for j in range(structure.groups_per_chain[i]):
                groupIndex = structure.group_type_list[groupCounter]
                atomsPerChain[i] += len(structure.group_list[groupIndex]['atomNameList'])
                bondsPerChain[i] += len(structure.group_list[groupIndex]['bondOrderList'])
                groupCounter += 1

        return atomsPerChain, bondsPerChain


    def _get_chain_to_entity_index(self, structure):
        '''Returns an list that maps a chain index to an entity index.

        Attributes
        ----------
            structure: structureDataInterFace
        '''
        entityChainIndex = [0] * structure.num_chains

        for i in range(len(structure.entity_list)):

            for j in structure.entity_list[i]["chainIndexList"]:
                entityChainIndex[j] = i

        return entityChainIndex



#!/user/bin/env python
'''experimentalMethods.py:

This filter returns ture if any of the specified experimental methods
match a PDB entry

References
----------
    The current list of support experimental method types can be found here:
        http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v40.dic/Items/_exptl.method.html

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''


class ExperimentalMethods(object):
    '''This filter returns True if any of the specified experimental methods
    matched a PDB entry.

    Attributes
    ----------
        experimental_methods (list(string)): A list of experimental methods to check
    '''

    # constants to be used as arguments to the Experimental Methods filter
    ELECTRON_CRYSTALLOGRAPHY = "ELECTRON CRYSTALLOGRAPHY"
    ELECTRON_MICROSCOPY = "ELECTRON MICROSCOPY"
    ERP = "EPR"
    FIBER_DIFFRACTION = "FIBER DIFFRACTION"
    FLUORESCENCE_TRANSFER = "FLUORESCENCE TRANSFER"
    INFRARED_SPECTROSCOPY = "INFRARED SPECTROSCOPY"
    NEUTRON_DIFFRACTION = "NEUTRON DIFFRACTION"
    POWDER_DIFFRACTION = "POWDER DIFFRACTION"
    SOLID_STATE_NMR = "SOLID-STATE NMR"
    SOLUTION_NMR = "SOLUTION NMR"
    SOLUTION_SCATTERING = "SOLUTION SCATTERING"
    THEORETICAL_MODEL = "THEORETICAL MODEL"
    X_RAY_DIFFRACTION = "X-RAY DIFFRACTION"

    def __init__(self, *experimentalMethods):
        self.experimental_methods = experimentalMethods

    def __call__(self, t):
        structure = t[1]
        methods = [b.upper() for b in structure.experimental_methods]
        return sum([1 for m in self.experimental_methods if m in methods]) > 0



'''sequenceNgrammer.py

This class contians methods for creating overlapping and non-overlapping
n-grams of one-letter code sequence
(e.g., protein sequences)

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from pyspark.sql import SparkSession, types


def ngram(data, n, outputCol):
    '''Splits a one-letter sequence column (e.g., protein sequence)
    into array of overlapping n-grams.

    Example
    -------
        2-gram: IDCGH ... => [ID, DC, CG, GH, ...]

    Attribute
    ---------
        data (dataset): input dataset with column "sequence"
        n (int): size of the n-gram
        outputCol (String): name of the output column

    Returns
    -------
        output dataset with appended ngram column
    '''

    session = SparkSession.builder.getOrCreate()

    #Encoder function to be passed as User Defined Function (UDF)
    def _ngrammer(s):
        ngram = []
        i = 0

        if len(s) < 1:
            return []

        while i < len(s) - n + 1:
            ngram.append(s[i: i + n])
            i += 1

        return ngram

    session.udf.register("ngrammer", _ngrammer, types.ArrayType(types.StringType()))

    data.createOrReplaceTempView("table")
    sql = f"SELECT *, ngrammer(sequence) AS {outputCol} from table"

    data = session.sql(sql)

    return data


def shifted_ngram(data, n, shift, outputCol):
    '''Splits a one-letter sequence column (e.g., protein sequence)
	into array of non-overlapping n-grams. To generate all possible n-grams,
	this method needs to be called n times with shift parameters {0, ..., n-1}.

	Example
    -------
        3-gram(shift=0) : IDCGHTVEDQR ... => [IDC, GHT, VED, ...]
	    3-gram(shift=1) : IDCGHTVEDQR ... => [DCG, HTV, EDQ, ...]
	    3-gram(shift=2) : IDCGHTVEDQR ... => [CGH, TVE, DQR, ...]

    References
    ----------
	   For anapplication of shifted n-grams see:
	   E Asgari, MRK Mofrad, PLoS One. 2015; 10(11): e0141287, doi:
	   https://dx.doi.org/10.1371/journal.pone.0141287

    Attribute
    ---------
        data (dataset): input dataset with column "sequence"
        n (int): size of the n-gram
        shift(int): start index for the n-gram
        outputCol (String): name of the output column

    Returns
    -------
        output dataset with appended ngram column
    '''

    session = SparkSession.builder.getOrCreate()

    #Encoder function to be passed as User Defined Function (UDF)
    def _ngrammer(s):
        ngram = []
        i,j = 0,0
        t = int(len(s)/n)

        if len(s) < shift:
            return []

        s = s[shift:]

        while j < t:
            ngram.append(s[i: i + n])
            j += 1
            i += n

        return ngram

    session.udf.register("ngrammer", _ngrammer, types.ArrayType(types.StringType()))

    data.createOrReplaceTempView("table")
    sql = f"SELECT *, ngrammer(sequence) AS {outputCol} from table"

    data = session.sql(sql)

    return data



#!/user/bin/env python
'''advancedQuery.py

This filter runs an RCSB PDB Advanced Search web service using an XML query
description.

References
----------
    Advanced Search Query
        https://www.rcsb.org/pdb/staticHelp.do?p=help/advancedSearch.html

Example
-------
    Find PDB entries that contain the word "mutant" in the structure title:

        query = "<orgPdbQuery>" + \
                "<queryType>org.pdb.query.simple.StructTitleQuery</queryType>" + \
                "<struct.title.comparator>contains</struct.title.comparator>" + \
                "<struct.title.value>mutant</struct.title.value" + \
                "</orgPdbQuery>"
        pdb = pdb.filter(AdvancedSearch(query));

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from mmtfPyspark.webservices.advancedQueryService import post_query


class AdvancedQuery(object):
    '''Filters using the RCSB PDB Advanced Search web service

    Attribute
    ---------
        xmlQuery (String): query in RCSB PDB XML format
    '''

    def __init__(self, xmlQuery):

        results = post_query(xmlQuery)

        self.entityLevel = (len(results) > 0) and (":" in results[0])
        self.structureIds = list(set(results))
        self.exclusive = False

    def __call__(self, t):

        structure = t[1]

        globalMatch = False
        numChains = structure.chains_per_model[0]
        entityChainIndex = self._get_chain_to_entity_index(structure)

        for i in range(numChains):

            ID = t[0]

            if self.entityLevel:
                ID = self._get_structure_entity_id(
                    structure, ID, entityChainIndex[i])

            match = ID in self.structureIds

            if match and not self.exclusive:
                return True

            if not match and self.exclusive:
                return False

            if match:
                globalMatch = True

        return globalMatch

    def _get_structure_entity_id(self, structure, origStructureId, origEntityId):

        keyStructureId = origStructureId

        try:
            index = keyStructureId.index(".")
            keyStructureId = keyStructureId[:index]
        except:
            pass

        try:
            pos = structure.structure_id.rindex(".")

            valueStructureId = structure.structure_id[:structure.structure_id.index(
                ".")]

            if keyStructureId != valueStructureId:
                raise Exception("Structure mismatch: key vs value: %s vs. %s"
                                % (keyStructureId, valueStructureId))

            entityId = structure.structure_id[pos + 1:]
            ID = valueStructureId + ":" + entityId

        except:
            ID = keyStructureId + ":" + str(origEntityId + 1)

        return ID

    def _get_chain_to_entity_index(self, structure):

        entityChainIndex = [0] * structure.num_chains

        for i in range(len(structure.entity_list)):

            for j in structure.entity_list[i]['chainIndexList']:

                entityChainIndex[j] = i

        return entityChainIndex



#!/user/bin/env python
'''customReportService.py

This class uses RCSB PDB Tabular Report RESTful webservices to retrieve
metadata and annotations for all current entries in the ProteinDataBank.

References
----------
    List of supported fieldnames:
        http://www.rcsb.org/pdb/results/reportField.do
    The RCSB Protein Data Bank:redesignedwebsiteandwebservices2011NucleicAcidsRes.39:D392-D401.
        https://dx.doi.org/10.1093/nar/gkq1021

Example
-------
    Retrieve PubMedCentral, PubMedID, and Depositiondate:
        ds = CustomReportService.getDataset("pmc","pubmedId","depositionDate")
        ds.printSchema()
        ds.show(5)

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import urllib
from urllib import request
import tempfile
from pyspark.sql import SparkSession


SERVICELOCATIONS = ["https://www.rcsb.org/pdb/rest/customReport",
                    "https://www1.rcsb.org/pdb/rest/customReport",
                    "https://www2.rcsb.org/pdb/rest/customReport"]


CURRENT_URL = "?pdbids=*&service=wsfile&format=csv&primaryOnly=1&customReportColumns="


def get_dataset(columnNames):
    '''Returns a dataset with the specified columns for all current PDB entires.
    See <a href="https://www.rcsb.org/pdb/results/reportField.do">  for a list
    of supported filed names

    Attributes
    ----------
        columnNames: names of columns for the dataset

    Returns
    -------
        dataset with the specified columns
    '''

    if type(columnNames) == str:
        columnNames = [columnNames]

    query = CURRENT_URL + ','.join(columnNames)

    dataset = _get_dataset(query, columnNames)

    return dataset


def _get_dataset(query, columnNames):
    '''Get dataset using different service locations
    '''

    dataset = None

    for SERVICELOCATION in SERVICELOCATIONS:
        while True:
            try:
                inStream = _post_query(SERVICELOCATION, query)

                tmp = tempfile.NamedTemporaryFile(delete=False)

                with open(tmp.name, "w") as t:
                    for l in inStream:
                        t.writelines(str(l)[2:-3] + '\n')

                spark = SparkSession.builder.getOrCreate()

                dataset = _read_csv(spark, tmp.name)
            except:
                continue
            break

    if dataset is None:
        print("ERROR: cannot connect to service location")
        return dataset

    return _concat_ids(spark, dataset, columnNames)


def _concat_ids(spark, dataset, columnNames):
    '''Concatenates structureId and chainId fields into a single key if chainId
    field is present

    Attributes
    ----------
        spark (SparkSession)
        dataset (Dataframe)
        columnNames (list): columnNames
    '''

    if "chainId" in dataset.columns:
        dataset.createOrReplaceTempView("table")

        sql = "SELECT CONCAT(structureId,'.',chainId) as structureChainId," + \
              "structureId,chainId,%s" % ','.join(columnNames) + \
              " from table"

        dataset = spark.sql(sql)

    return dataset


def _post_query(service, query):
    '''Post PDB Ids and fields in a query string to the RESTful RCSB web service

    Attributes
    ----------
        query (string): RESTful query urlopen
        service (str): service location

    Returns
    -------
        input stream to response
    '''

    encodedQuery = urllib.parse.quote(query).encode('utf-8')
    url = request.Request(service)
    stream = urllib.request.urlopen(url, data=encodedQuery)

    return stream


def _read_csv(spark, inputFileName):
    '''Reads CSV file into a Spark dataset

    Attributes
    ----------
        spark (Spark Context)
        inputFileName (str): directory path for the input file
    '''

    dataset = spark.read \
                   .format("csv") \
                   .option("header", "true") \
                   .option("inferSchema", "true") \
                   .load(inputFileName) \
                   .cache()

    return dataset



#!/user/bin/env python
'''distanceBox.py:

This code is a modification from BioJava's distanceBox class

Reference
---------
    BioJava distanceBox:
        https://github.com/biojava/biojava/blob/359dce819e7678879ae41301b11fa268f92e81d1/biojava-structure/src/main/java/org/biojava/nbio/structure/symmetry/geometry/DistanceBox.java

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''
import numpy as np
from collections import defaultdict


class DistanceBox(object):

    offset = [
        0 + (0 * 10000) + (0 * 1000000000),
        -1 + (-1 * 10000) + (-1 * 1000000000),
        -1 + (-1 * 10000) + (0 * 1000000000),
        -1 + (-1 * 10000) + (1 * 1000000000),
        -1 + (0 * 10000) + (-1 * 1000000000),
        -1 + (0 * 10000) + (0 * 1000000000),
        -1 + (0 * 10000) + (1 * 1000000000),
        -1 + (1 * 10000) + (-1 * 1000000000),
        -1 + (1 * 10000) + (0 * 1000000000),
        -1 + (1 * 10000) + (1 * 1000000000),
        0 + (-1 * 10000) + (-1 * 1000000000),
        0 + (-1 * 10000) + (0 * 1000000000),
        0 + (-1 * 10000) + (1 * 1000000000),
        0 + (0 * 10000) + (-1 * 1000000000),
        0 + (0 * 10000) + (1 * 1000000000),
        0 + (1 * 10000) + (-1 * 1000000000),
        0 + (1 * 10000) + (0 * 1000000000),
        0 + (1 * 10000) + (1 * 1000000000),
        1 + (-1 * 10000) + (-1 * 1000000000),
        1 + (-1 * 10000) + (0 * 1000000000),
        1 + (-1 * 10000) + (1 * 1000000000),
        1 + (0 * 10000) + (-1 * 1000000000),
        1 + (0 * 10000) + (0 * 1000000000),
        1 + (0 * 10000) + (1 * 1000000000),
        1 + (1 * 10000) + (-1 * 1000000000),
        1 + (1 * 10000) + (0 * 1000000000),
        1 + (1 * 10000) + (1 * 1000000000)
    ]

    def __init__(self, binWidth):

        self.inverseBinWidth = 1.0 / binWidth
        self.hashMap = defaultdict(list)

    def add_point(self, point, pointName):

        i = np.rint(float(point[0]) * self.inverseBinWidth)
        j = np.rint(float(point[1]) * self.inverseBinWidth)
        k = np.rint(float(point[2]) * self.inverseBinWidth)
        location = i + (j * 10000) + (k * 1000000000)

        self.hashMap[location].append(pointName)

    def get_neighbors(self, point):

        i = np.rint(float(point[0]) * self.inverseBinWidth)
        j = np.rint(float(point[1]) * self.inverseBinWidth)
        k = np.rint(float(point[2]) * self.inverseBinWidth)
        location = i + (j * 10000) + (k * 1000000000)

        box = self.get_box_two(location)

        return box

    def get_box_two(self, location):

        box_two = []
        for off in self.offset:
            if (location + off) in self.hashMap:
                box_two += self.hashMap[location + off]

        return box_two
        '''
        return [self.hashMap[location + off] for off in self.offset
                if (location + off) in self.hashMap]
        '''
    def getIntersection(self, distanceBox):

        intersection = []
        checkedLocations = set()

        for location in self.hashMap.keys():
            overlap = False

            for off in self.offset:
                loc = location + off

                if loc in distanceBox.hashMap:

                    overlap = True
                    break

            if overlap:
                for off in self.offset:
                    loc = location + off

                    if loc in checkedLocations:
                        continue

                    checkedLocations.add(loc)
                    if loc in self.hashMap:
                        intersection += self.hashMap[loc]

        return intersection



#!/usr/bin/env python
'''datasetRegressor.py

Runs regression on a given dataset.
Dataset are read as Parquet file. The dataset must contain
a feature vector named "features" and a prediction column.
The column name of the prediction column must be specified
on the command lines.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.1"
    __status__ = "Done"
'''

from mmtfPyspark.ml import SparkRegressor, datasetBalancer
from pyspark.sql import SparkSession
from pyspark.ml.regression import GBTRegressor, GeneralizedLinearRegression, LinearRegression
import sys
import time

def main(argv):

    # Name of prediction column
    label = argv[1]

    start = time.time()

    spark = SparkSession.builder \
                        .master("local[*]") \
                        .appName("datasetRegressor") \
                        .getOrCreate()

    data = spark.read.parquet(argv[0]).cache()

    vector = data.first()
    print(vector)
    featureCount = len(vector)
    print("Feature count    : {featureCount}")

    print("Dataset size (unbalanced)    : {data.count()}")

    testFraction = 0.3
    seed = 123

    # Linear Regression
    lr = LinearRegression().setLabelCol(label) \
                           .setFeaturesCol("features")
    reg = SparkRegressor(lr, label, testFraction, seed)
    matrics = reg.fit(data)
    for k,v in matrics.items(): print(f"{k}\t{v}")

    # GBTRegressor
    gbt = GBTRegressor().setLabelCol(label) \
                        .setFeaturesCol("features")
    reg = SparkRegressor(gbt, label, testFraction, seed)
    matrics = reg.fit(data)
    for k,v in matrics.items(): print(f"{k}\t{v}")

    # GeneralizedLinearRegression
    glr = GeneralizedLinearRegression().setLabelCol(label) \
                                       .setFeaturesCol("features") \
                                       .setFamily("gaussian") \
                                       .setLink("identity") \
                                       .setMaxIter(10) \
                                       .setRegParam(0.3)
    reg = SparkRegressor(glr, label, testFraction, seed)
    matrics = reg.fit(data)
    for k,v in matrics.items(): print(f"{k}\t{v}")

    end = time.time()
    print("Time: %f  sec." %(end-start))


if __name__ == "__main__":

    if len(sys.argv) < 3:
        raise Exception("python datasetClassifier.py <parquet file> <prediction column name>")
        sys.exit()

    main(sys.argv[1:])



#!/user/bin/env python
'''pythonRDDToDataset.py:

This class converts a PythonRDD<Row> to a Dataset<Row>. This method only
supports simple data types and all data need to be not null.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from pyspark.sql.types import *
from pyspark.sql import SparkSession

def get_dataset(data, colNames):
    '''Converts a PythonRDD<Row> to a Dataset<Row>. This method only
	supports simple data types and all data need to be not null.

    Attributes
    ----------
        data (PythonRDD): PythonRDD of row objects
        colNames (list(str)): names of the columns in a row
    '''

    row = data.first()
    length = len(row)

    if length != len(colNames):
        raise Exception("colNames length does not match row length")

    sf = []

    for i in range(len(colNames)):
        o = row[i]
        if type(o) == str:
            sf.append(StructField(colNames[i], StringType(), False))
        elif type(o) == int:
            sf.append(StructField(colNames[i], IntegerType(), False))
        elif type(o) == float:
            sf.append(StructField(colNames[i], FloatType(), False))
        elif type(o) == long:
            sf.append(StructField(colNames[i], LongType(), False))
        else:
            print("Data type not implemented yet")

    schema = StructType(sf)
    spark = SparkSession.builder.getOrCreate()
    return spark.createDataFrame(data, schema)



#!/user/bin/env python
'''interactionFingerprinter.py

This class creates dataset of ligand - macromolecule and macromolecule -
macromolecule interaction information. Criteria to select interactions are
specified by the InteractionFilter.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

from pyspark.sql import SparkSession
from pyspark.sql.types import *
from pyspark import SparkContext
from mmtfPyspark.interactions import LigandInteractionFingerprint, PolymerInteractionFingerprint


class InteractionFingerprinter(object):

    def get_ligand_polymer_interactions(structures, interactionFilter):
        '''Returns a dataset of ligand - macromolecule interacting residues.

        The dataset contains the following columns:
            structureChainId - pdbId.chainName of chain that interacts with ligand
            queryLigandId - id of ligand from PDB chemical component dictionary
            queryLigandNumber - group number of ligand including insetion code
            queryLigandChainId - chain name of ligand
            targetChainId - name of chain for which the interaction data are listed
            groupNumbers - array of residue number of interacting groups including insertion code (e.g. 101A)
            sequenceIndices - array of zero-based index of interaction groups (residues) mapped onto target sequence
            sequence - interacting polymer sequence
            interactingChains - total number of chains that interact with ligand

        Attributes
        ----------
            structures (PythonRDD): a set of PDB structures
            interactionFilter (InteractionFilter): interaction criteria

        Returns
        -------
            dataset with interacting residue information
        '''

        # find sll interactions
        row = structures.flatMap(LigandInteractionFingerprint(interactionFilter))

        # convert RDD to a Dataset with the following columns
        nullable = False
        fields = [StructField("structureChainId", StringType(), nullable),
                  StructField("queryLigandId", StringType(), nullable),
                  StructField("queryLigandNumber", StringType(), nullable),
                  StructField("queryLigandChainId", StringType(), nullable),
                  StructField("targetChainId", StringType(), nullable),
                  StructField("groupNumbers", ArrayType(StringType(), nullable), nullable),
                  StructField("sequenceIndices", ArrayType(IntegerType(), nullable), nullable),
                  StructField("sequence", StringType(), nullable),
                  StructField("interactingChains", IntegerType(), nullable)
                  ]

        schema = StructType(fields)
        spark = SparkSession.builder.getOrCreate()
        return spark.createDataFrame(row, schema)


    def get_polymer_interactions(structures, interactionFilter):
        '''Returns a dataset of ligand - macromolecule interacting information.

        The dataset contains the following columns:
            structureChainId - pdbId.chainName for which the interaction data are listed
            queryChainId - name of chain that interacts with target chain
            targetChainId - name of chain for which the interaction data are listed
            groupNumbers - array of residue number of interacting groups including insertion code (e.g. 101A)
            sequenceIndices - array of zero-based index of interaction groups (residues) mapped onto target sequence
            sequence - target polymer sequence

        Attributes
        ----------
            structures (PythonRDD): a set of PDB structures
            interactionFilter (InteractionFilter): interaction criteria

        Returns
        -------
            dataset with interacting residue information
        '''

        # find all interactions
        row = structures.flatMap(PolymerInteractionFingerprint(interactionFilter))

        # convert RDD to a Dataset with the following columns
        nullable = False
        fields = [StructField("structureChainId", StringType(), nullable),
                  StructField("queryChainId", StringType(), nullable),
                  StructField("targetChainId", StringType(), nullable),
                  StructField("groupNumbers", ArrayType(StringType(),nullable), nullable),
                  StructField("sequenceIndices", ArrayType(IntegerType(), nullable), nullable),
                  StructField("sequence", StringType(), nullable),
                  ]

        schema = StructType(fields)
        spark = SparkSession.builder.getOrCreate()
        return spark.createDataFrame(row, schema)



'''uniProt.py

This class downloads and reads UniProt sequence files in the FASTA format and
converts them to datasets.This module reads the following files:
    SWISS_PROT,
    TREMBL,
    UNIREF50,
    UNIREF90,
    UNIREF100.

Reference
---------
    UniProt downloads:
        http://www.uniprot.org/downloads
    The datasets have the following columns:
        http://www.uniprot.org/help/fasta-headers

Example
-------
    Download, read, and save the SWISS_PROT dataset:
        ds = uniProt.get_dataset(UniProtDataset.SWISS_PROT)
        ds.printSchema()
        ds.show(5)
        ds.write().mode("overwrite").format("parquet").save(fileName)

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from collections import namedtuple
import tempfile
import urllib.request
import gzip
from pyspark.sql import SparkSession
from enum import Enum


baseUrl = "ftp://ftp.uniprot.org/pub/databases/uniprot/"

SWISS_PROT = baseUrl + "current_release/knowledgebase/complete/uniprot_sprot.fasta.gz"
TREMBL = baseUrl + "current_release/knowledgebase/complete/uniprot_trembl.fasta.gz"
UNIREF50 = baseUrl + "uniref/uniref50/uniref50.fasta.gz"
UNIREF90 = baseUrl + "uniref/uniref90/uniref90.fasta.gz"
UNIREF100 = baseUrl + "uniref/uniref100/uniref100.fasta.gz"


def _get_uniprot_dataset(dataType):
    '''
    Get Uniprot Dataset
    '''

    # Decalre string variables
    db, uniqueIdentifier, entryName, proteinName, organismName, geneName, \
        proteinExistence, sequenceVersion, sequence = '', '', '', '', '', '', '', '', ''
    firstLine = True

    # Make temporary file
    tempFile = tempfile.NamedTemporaryFile(delete=False)
    with open(tempFile.name, "w") as t:

        t.writelines(
            "db,uniqueIdentifier,entryName,proteinName,organismName,geneName,proteinExistence,sequenceVersion,sequence\n")

        inputStream = urllib.request.urlopen(dataType, timeout=6000)
        rd = gzip.GzipFile(fileobj=inputStream)

        for line in rd:  # TODO check rd output content after UNIPROT online

            line = str(line)[2:-3]

            if ">" in line:
                line = line.replace(",", ";")

                if not firstLine:
                    t.writelines(
                        f"{db},{uniqueIdentifier},{entryName},{proteinName},{organismName},{geneName},{proteinExistence},{sequenceVersion}, {sequence}\n".replace(' ', ''))

                firstLine = False
                sequence = ""
                tmp = line.split("|")  # TODO
                db = tmp[0]
                uniqueIdentifier = tmp[1]
                tmp = tmp[2]  # TODO

                if len(tmp.split(" OS=")) > 2:
                    length = len(tmp.split(" OS")[0]) + \
                        len(tmp.split(" OS")[1]) + 4
                    tmp = tmp[:length]

                # Set sequence version
                sv = tmp.split(" SV=")
                tmp = sv[0]
                sequenceVersion = sv[1] if len(sv) > 1 else ""

                # Set proteinExistence
                pe = tmp.split(" PE=")
                tmp = pe[0]
                proteinExistence = pe[1] if len(pe) > 1 else ""

                # Set GeneName
                ge = tmp.split(" GN=")
                tmp = ge[0]
                geneName = ge[1] if len(ge) > 1 else ""

                # Set organismName
                on = tmp.split(" OS=")
                tmp = on[0]
                organismName = on[1] if len(on) > 1 else ""

                entryName = tmp.split(" ")[0]

                proteinName = tmp[len(entryName) + 1:]
            else:
                sequence += line

    spark = SparkSession.builder.getOrCreate()

    dataset = spark.read \
                   .format("csv") \
                   .option("header", "true") \
                   .option("inferSchema", "true") \
                   .load(tempFile.name)
    return dataset


def _get_uniref_dataset(dataType):
    '''
    Get Uniref Dataset
    '''

    # Decalre string variables
    uniqueIdentifier, clusterName, taxon, representativeMember, taxonID, \
        members = '', '', '', '', '', ''
    firstLine = True

    # Make temporary file
    tempFile = tempfile.NamedTemporaryFile(delete=False)
    with open(tempFile.name, "w") as t:

        t.writelines(
            "uniqueIdentifier,clusterName,members,taxon,taxonID,representativeMember,sequence\n")

        inputStream = urllib.request.urlopen(dataType)
        rd = gzip.GzipFile(fileobj=inputStream)

        for line in rd:  # TODO check rd output content after UNIPROT online

            line = str(line)[2:-3]

            if ">" in line:

                line = line.replace(",", ";")

                if not firstLine:
                    t.writelines(
                        f"{uniqueIdentifier},{clusterName},{members},{taxon},{taxonID},{representativeMember},{sequence}\n".replace(' ', ''))

                firstLine = False

                sequence = ""
                tmp = line

                # Set representativeMember
                rm = tmp.split(" RepID=")
                tmp = rm[0]
                representativeMember = rm[1] if len(rm) > 1 else ""

                # Set taxonID
                tid = tmp.split(" TaxID=")
                tmp = tid[0]
                taxonID = tid[1] if len(tid) > 1 else ""

                # Set taxon
                tx = tmp.split(" Tax=")
                tmp = tx[0]
                taxon = tx[1] if len(tx) > 1 else ""

                # Set members
                m = tmp.split(" n=")
                tmp = m[0]
                members = m[1] if len(m) > 1 else ""

                uniqueIdentifier = tmp.split(" ")[0]
                clusterName = tmp[len(uniqueIdentifier) + 1:]
            else:
                sequence += line

    spark = SparkSession.builder.getOrCreate()

    dataset = spark.read \
                   .format("csv") \
                   .option("header", "true") \
                   .option("inferSchema", "true") \
                   .load(tempFile.name)
    return dataset


def get_dataset(UniProtDataset):
    '''Returns the specified UniProt dataset.

    Attributes
    ----------
        uniProtDataset (String): name of the UniProt dataset

    Returns
    -------
        dataset with sequence and metadata
    '''

    if UniProtDataset.split('/')[-3] == "uniref":

        return _get_uniref_dataset(UniProtDataset)

    elif UniProtDataset.split('/')[-3] == "knowledgebase":

        return _get_uniprot_dataset(UniProtDataset)

    else:

        raise Exception("Please use pre-defined uniprotDataset \n \
                         eg: UniprotDataset.SWISS_PROT")



#!/user/bin/env python
'''containsPolymerChainType.py

This filter returns entries that contain chains made of the specified
monomer types. The default constructor returns entries that contain at least
one chain that matches the conditions. If the "exclusive" flag is set to true
in the constructor, all chains must match the conditions. For a multi-model
structure, this filter only checks the first model.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''


class ContainsPolymerChainType(object):
    '''Default constructor matches any entry that contains a chain with only
    the specified monomer type

    Attributes
    ----------
        monomer_type (list(str)): list of monomer types in a polymer chain
    '''

    D_PEPTIDE_COOH_CARBOXY_TERMINUS = "D-PEPTIDE COOH CARBOXY TERMINUS"
    D_PEPTIDE_NH3_AMINO_TERMINUS = "D-PEPTIDE NH3 AMINO TERMINUS"
    D_PEPTIDE_LINKING = "D-PEPTIDE LINKING"
    D_SACCHARIDE = "D-SACCHARIDE"
    D_SACCHARIDE_14_and_14_LINKING = "D-SACCHARIDE 1,4 AND 1,4 LINKING"
    D_SACCHARIDE_14_and_16_LINKING = "D-SACCHARIDE 1,4 AND 1,6 LINKING"
    DNA_OH_3_PRIME_TERMINUS = "DNA OH 3 PRIME TERMINUS"
    DNA_OH_5_PRIME_TERMINUS = "DNA OH 5 PRIME TERMINUS"
    DNA_LINKING = "DNA LINKING"
    L_PEPTIDE_COOH_CARBOXY_TERMINUS = "L-PEPTIDE COOH CARBOXY TERMINUS"
    L_PEPTIDE_NH3_AMINO_TERMINUS = "L-PEPTIDE NH3 AMINO TERMINUS"
    L_PEPTIDE_LINKING = "L-PEPTIDE LINKING"
    L_SACCHARIDE = "L-SACCHARIDE"
    L_SACCHARIDE_14_AND_14_LINKING = "L-SACCHARDIE 1,4 AND 1,4 LINKING"
    L_SACCHARIDE_14_AND_16_LINKING = "L-SACCHARIDE 1,4 AND 1,6 LINKING"
    PEPTIDE_LINKING = "PEPTIDE LINKING"
    RNA_OH_3_PRIME_TERMINUS = "RNA OH 3 PRIME TERMINUS"
    RNA_OH_5_PRIME_TERMINUS = "RNA OH 5 PRIME TERMINUS"
    RNA_LINKING = "RNA LINKING"
    NON_POLYMER = "NON-POLYMER"
    OTHER = "OTHER"
    SACCHARIDE = "SACCHARIDE"

    def __init__(self, monomer_type, exclusive=False):
        if type(monomer_type) == str:
            monomer_type = monomer_type.split(',')

        self.exclusive = exclusive
        self.monomer_type = monomer_type

    def __call__(self, t):
        structure = t[1]
        contrains_polymer = False
        global_match = False
        # get number of chains in first model, nessary?
        num_chains = structure.chains_per_model[0]
        group_counter = 0

        for i in range(num_chains):
            match = True
            chain_type = [chain['type'] for chain in structure.entity_list
                          if i in chain['chainIndexList']][0]
            polymer = chain_type == "polymer"

            if polymer:
                contains_polymer = True
            else:
                match = False

            for j in range(structure.groups_per_chain[i]):
                if match and polymer:
                    group_idx = structure.group_type_list[group_counter]
                    group_type = structure.group_list[group_idx]['chemCompType']
                    match = (group_type in self.monomer_type)
                group_counter += 1

            if (polymer and match and not self.exclusive):
                return True

            if (polymer and not match and self.exclusive):
                return False

            if match:
                global_match = True

        return global_match and contains_polymer



from .interactionFilter import InteractionFilter
from .interactionCenter import InteractionCenter
from .coordinationGeometry import CoordinateGeometry
from .atomInteraction import AtomInteraction
from .structureToAtomInteractions import StructureToAtomInteractions
from .groupInteractionExtractor import GroupInteractionExtractor
from .ligandInteractionFingerprint import LigandInteractionFingerprint
from .polymerInteractionFingerprint import PolymerInteractionFingerprint
from .interactionFingerprinter import InteractionFingerprinter



#!/user/bin/env python
'''polymerSequenceExtractor.py:

Creates a dataset of polymer sequences using the full sequence
used in the experiment (i.e., the "SEQRES" record in PDB files).

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Debug"
'''

from mmtfPyspark.ml import pythonRDDToDataset
from mmtfPyspark.mappers import StructureToPolymerSequences
from pyspark.sql import Row


def get_dataset(structures):
    '''Returns a dataset of polymer sequence contained in PDB entries
    using the full sequence used in the experimnet
    (i.e., the "SEQRES" record in PDB files)

    Attributes
    ----------
        structures (pythonRDD): a set of PDB structures

    Returns
    -------
        dataset with interacting residue and atom information
    '''

    rows = structures.flatMap(StructureToPolymerSequences()) \
                     .map(lambda x: Row(x[0],x[1]))

    colNames = ["structureChainId", "sequence"]

    return pythonRDDToDataset.get_dataset(rows, colNames)



from .structureToPolymerChains import StructureToPolymerChains
from .structureToPolymerSequences import StructureToPolymerSequences
from .structureToBioassembly import StructureToBioassembly
from .structureToSecondaryStructureSegments import StructureToSecondaryStructureSegments
from .structureToSecondaryStructureElements import StructureToSecondaryStructureElements
from .structureToProteinDimers import StructureToProteinDimers
from .structureToBiopython import StructureToBiopython
from .structureToInteractingResidues import StructureToInteractingResidues



#!/user/bin/env python
'''interactionFilter.py

A filter to specify criteria for molecular interactions between a query and a
target within a macromolecular structure. The filter specifies criteria for the
query (e.g. a metal ion) and the target (e.g. amino acid redisudes).
Interaction criteria such as distance cutoff limit the nature of interactions to
be considered.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

import sys


class InteractionFilter(object):
    ''' A filter to specify criteria for molecular interactions between a query
    and a target

    Attributes
    ----------
        distanceCutoff (float): cutoff distance for filter [float max]
        normalizedbFactorCutoff (float) normalized bFactor cutoff for filter [float max]
        minInteractions (int): minimum number of interactions [1]
        maxInteractions (int): maximum number of interactions [10]
    '''

    def __init__(self, distanceCutoff=sys.float_info.max -1,
                 normalizedbFactorCutoff=sys.float_info.max -1,
                 minInteractions=1, maxInteractions=10):

        # Set interaction criteria
        self._distanceCutoff = distanceCutoff
        self._normalizedbFactorCutoff = normalizedbFactorCutoff
        self._minInteractions = minInteractions
        self._maxInteractions = maxInteractions

        # Query criteria
        self._queryGroups = None
        self._queryElements = None
        self._queryAtomNames = None

        # Target criteria
        self._targetGroups = None
        self._targetElements = None
        self._targetAtomNames = None
        self._prohibitedTargetGroups = None

    def get_distance_cutoff(self):
        '''Get the maximum interaction distance. At least one pair of query and
        target atoms must be wintin in distance cutoff

        Returns
        -------
            maximum interaction distance
        '''
        return self._distanceCutoff

    def set_distance_cutoff(self, distanceCutoff):
        '''Set the distnace cutoff for interacting atoms

        Attributes
        ----------
            distanceCutoff (float): the maximum distance for interacting atoms
        '''

        self._distanceCutoff = distanceCutoff

    def get_normalized_b_factor_cutoff(self):
        '''Gets the maximum normalized b-factor (z-score) cutoff for an atom and
        its interacting neighbor atoms.

        Returns
        -------
            maximum normalized b-factor cutoff
        '''

        return self._normalizedbFactorCutoff

    def set_normalized_b_factor_cutoff(self, normalizedbFactorCutoff):
        '''Sets the maximum noramlized b-factor cutoff. This value represents a
        z-score (see reference below), the signed number of standard deviations
        by which the b-factor. High z-scores indicate either high flexibility
        and/or experimental error for the atoms involved in the interactions. By
        setting a cutoff value, not well defined interactions can be exculded.

        Frequently used z-scores:

        Confidence level    Tail area   z-scores
                     90%        0.05     +-1.645
                     95%        0.025    +-1.96
                     99%        0.005    +-2.576

        For example, to include all interactions within the 90% confidence
        interval, set the normalized b-factor to +1.645

        Attribute
        ---------
            normalizedbFactorCutoff (float): maximum normalized b-factor

        Reference
        ---------
            Z-score: https://en.wikipedia.org/wiki/Standard_score
        '''

        self._normalizedbFactorCutoff = normalizedbFactorCutoff

    def get_min_interactions(self):
        '''Returns the minimum number of interactions per atom. Atoms that interact
        with fewer atoms wil be discarded.

        Returns
        -------
            minimum number of interactions per atom
        '''

        return self._minInteractions

    def set_min_interactions(self, minInteractions):
        '''Sets the minimum number of interactions per atom. Atoms that interact
        with fewer atoms will be discarded

        Attribute
        ---------
            minInteractions (int): minimum number of interactions per atom
        '''

        self._minInteractions = minInteractions

    def get_max_interactions(self):
        '''Returns the maximum number of interactions per atom. Atoms that interact
        with fewer atoms wil be discarded.

        Returns
        -------
            maximum number of interactions per atom
        '''

        return self._maxInteractions

    def set_max_interactions(self, maxInteractions):
        '''Sets the maximum number of interactions per atom. Atoms that interact
        with fewer atoms will be discarded

        Attribute
        ---------
            maxInteractions (int): minimum number of interactions per atom
        '''

        self._maxInteractions = maxInteractions

    def set_query_elements(self, include, elements):
        '''Sets the elements to either be included or excluded in the query.
        Element strings are case sensitive (e.g., "Zn" for Zinc).

        Example: Only use elements O, N, S in the query groups in find polar
        interactions

            filter = InteractionFilter()
            filter.set_query_elements(True, ["O", "N", "S"])

        Example: Exclude non-polar elements and hydrogen in query groups and use
        all other elements to find interactions.

            elements = ['C', 'H', 'P']
            filter.set_query_elements(False, elements)

        Attributes
        ----------
            include (bool): if True, uses the specifed elements in the query,
                            if False, ignores the specified elemetns and use all
                            other elements
            elements (list): list of elements to be included or excluded in query
        '''

        if self._queryElements is not None:
            raise ValueError("ERROR: QueryElements have already been set.")

        if type(elements) == str:
            elements = [elements]
        self._includeQueryElements = include
        self._queryElements = set(elements)

    def set_target_elements(self, include, elements):
        '''Sets the elements to either be included or excluded in the target.
        Element strings are case sensitive (e.g., "Zn" for Zinc).

        Example: Only use elements O, N, S in the target groups in find polar
        interactions

            filter = InteractionFilter()
            filter.set_target_elements(True, ["O", "N", "S"])

        Example: Exclude non-polar elements and hydrogen in target groups and use
        all other elements to find interactions.

            elements = ['C', 'H', 'P']
            filter.set_query_elements(False, elements)

        Attributes
        ----------
            include (bool): if True, uses the specifed elements in the target,
                            if False, ignores the specified elemetns and use all
                            other elements
            elements (list): list of elements to be included or excluded in target
        '''

        if self._targetElements is not None:
            raise ValueError("ERROR: TargetElements have already been set.")

        if type(elements) == str:
            elements = [elements]
        self._includeTargetElements = include
        self._targetElements = set(elements)

    def set_query_groups(self, include, groups):
        '''Sets groups to either be included or excluded in the query. Group names
        must be upper case (e.g. 'ZN' for Zinc)

        Example: Find interactions with ATP and ADP

            filter = InteractionFilter()
            filter.set_query_groups(True, ['ATP', 'ADP'])

        Example: Exclude water and heavy water and use all other groups to find
        interactions.

            groups = ["HOH", "DOD"]
            filter.set_query_groups(False, groups)

        Attributes
        ----------
            include (bool): if True, uses the specified groups in the query,
                            if False, ignores the specified groups and uses all
                            other groups
            groups (list): groups to be included or excluded in query
        '''

        if self._queryGroups is not None:
            raise ValueError("ERROR: QueryGroups have already been set.")

        if type(groups) == str:
            groups = [groups]
        self._includeQueryGroups = include
        self._queryGroups = set(groups)

    def set_target_groups(self, include, groups):
        '''Sets groups to either be included or excluded in the target. Group names
        must be upper case (e.g. 'ZN' for Zinc)

        Example: Find interactions with specific amino acid groups.

            filter = InteractionFilter()
            filter.set_target_groups(True, ['CYS','HIS','ASP','GLU'])

        Example: Exclude water and heavy water and use all other groups to find
        interactions.

            groups = ["HOH", "DOD"]
            filter.set_target_groups(False, groups)

        Attributes
        ----------
            include (bool): if True, uses the specified groups in the target,
                            if False, ignores the specified groups and uses all
                            other groups
            groups (list): groups to be included or excluded in query
        '''

        if self._targetGroups is not None:
            raise ValueError("ERROR: QueryGroups have already been set.")

        if type(groups) == str:
            groups = [groups]
        self._includeTargetGroups = include
        self._targetGroups = set(groups)

    def set_query_atom_names(self, include, atomNames):
        '''Set atom names to either be included or excluded in the query

        Examples
        --------
            Find interaction with C-alpha and C-beta atoms.

                filter = InteractionFilter()
                filter.set_query_atom_names(True, ['CA', 'CB'])

            Exclude backbone atoms, but consider all other atom names: e.g,
            amino acid side chains.

                filter.set_query_atom_names(False, ['N', 'CA', 'C', 'O'])

        Attributes
        ----------
            include (bool): if True, uses set of atom names in query, if False,
                            ignores atoms with the specified names and uses all
                            other atoms
            atom_names (list): atoms to be included or excluded in the query
        '''

        if self._queryAtomNames is not None:
            raise ValueError("ERROR: QueryGroups have already been set.")

        if type(atomNames) == str:
            atomNames = [atomNames]
        self._includeQueryAtomNames = include
        self._queryAtomNames = set(atomNames)

    def set_target_atom_names(self, include, atomNames):
        '''Set atom names to either be included or excluded in the target.

        Examples
        --------
            Find interaction with C-alpha and C-beta atoms.

                filter = InteractionFilter()
                filter.set_target_atom_names(True, ['CA', 'CB'])

            Exclude backbone atoms, but consider all other atom names: e.g,
            amino acid side chains.

                filter.set_target_atom_names(False, ['N', 'CA', 'C', 'O'])

        Attributes
        ----------
            include (bool): if True, uses set of atom names in target, if False,
                            ignores atoms with the specified names and uses all
                            other atoms
            atom_names (list): atoms to be included or excluded in the target
        '''

        if self._targetAtomNames is not None:
            raise ValueError("ERROR: QueryGroups have already been set.")

        if type(atomNames) == str:
            atomNames = [atomNames]
        self._includeTargetAtomNames = include
        self._targetAtomNames = set(atomNames)

    def set_prohibited_target_groups(self, groups):
        '''Sets groups that must not appear in interactions. Any interactions that
        involves the specified groups will be excluded from the results.

        Example Find Zinc interactions, but discard any interactions where the
        metal is involved in an interaction with water.

            filter = InteractionFilter()
            filter.set_query_groups(True, 'ZN')
            filter.set_prohibited_target_groups(["HOH"])

        Attributes
        ----------
            groups(list): one or more group names to be prohibited
        '''
        if type(groups) == str:
            groups = [groups]
        self._prohibitedTargetGroups = set(groups)

    def is_query_element(self, element):
        '''Returns True if the specified elements matches the query conditions.

        Attributes
        ----------
            element (string): the element to be checked
        Returns
        -------
            True if element matches query conditinos, else False
        '''

        if self._queryElements is None:
            return True

        if self._includeQueryElements:
            return element in self._queryElements
        else:
            return element not in self._queryElements

    def is_target_element(self, element):
        '''Returns True if the specified elements matches the target conditions.

        Attributes
        ----------
            element (string): the element to be checked
        Returns
        -------
            True if element matches target conditinos, else False
        '''

        if self._targetElements is None:
            return True

        if self._includeTargetElements:
            return element in self._targetElements
        else:
            return element not in self._targetElements

    def is_query_group(self, group):
        '''Returns True if the specified group matches the query conditions.

        Attributes
        ----------
            group (string): the group to be checked
        Returns
        -------
            True if group matches query conditinos, else False
        '''

        if self._queryGroups is None:
            return True

        if self._includeQueryGroups:
            return group in self._queryGroups
        else:
            return group not in self._queryGroups

    def is_target_group(self, group):
        '''Returns True if the specified group matches the target conditions.

        Attributes
        ----------
            group (string): the group to be checked
        Returns
        -------
            True if group matches target conditinos, else False
        '''

        if self._targetGroups is None:
            return True

        if self._includeTargetGroups:
            return group in self._targetGroups
        else:
            return group not in self._targetGroups

    def is_query_atom_name(self, atomName):
        '''Returns True if the specified atom matches the query conditions.

        Attributes
        ----------
            atomName (string): the atom name to be checked
        Returns
        -------
            True if atom matches query conditinos, else False
        '''

        if self._queryAtomNames is None:
            return True

        if self._includeQueryAtomNames:
            return atomName in self._queryAtomNames
        else:
            return atomName not in self._queryAtomNames

    def is_target_atom_name(self, atomName):
        '''Returns True if the specified atom matches the target conditions.

        Attributes
        ----------
            atomName (string): the atom to be checked
        Returns
        -------
            True if atom matches target conditinos, else False
        '''

        if self._targetAtomNames is None:
            return True

        if self._includeTargetAtomNames:
            return atomName in self._targetAtomNames
        else:
            return atomName not in self._targetAtomNames

    def is_prohibited_target_group(self, group):
        '''Returns True if the specified group must not occur in an interactions.

        Attributes
        ----------
            group (str): group that must not occur in interactions
        Returns
        -------
            True if group is prohibited else False
        '''

        if self._prohibitedTargetGroups is None:
            return False

        return group in self._prohibitedTargetGroups



#!/user/bin/env python
'''secondaryStructure.py

This filter returns entries that contain polymer chain(s) with the specified
fraction of secondary structure assignments, obtained by DSSP. Note, DSSP
secondary structure in MMTF files is assigned by the BioJava implementation of
DSSP. It may differ in some cases from the original DSSP implementation.


Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''
from mmtfPyspark.utils import DsspSecondaryStructure


class SecondaryStructure(object):
    '''The default constructor returns entries that contain at least one
    polymer chain that matches the criteria. If the "exclusive" flag is set to
    true in the constructor, all polymer chains must match the criteria. For a multi-model
    structure, this filter only checks the first model.

    Attributes
    ----------
        helixFractionMin (float): minimum value for helix fraction [0.0]
        helixFractionMax (float): maximum value for helix fraction [1.0]
        sheetFractionMin (float): minimum value for sheet fraction [0.0]
        sheetFractionMax (float): maximum value for sheet fraction [1.0]
        coilFractionMin (float): minimum value for coil fractions [0.0]
        coilFractionMax (float): maximum value for coil fractions [1.0]
        exclusive (bool): exclusive flag [False]
    '''

    def __init__(self, helixFractionMin=0.0, helixFractionMax=1.0,
                 sheetFractionMin=0.0, sheetFractionMax=1.0,
                 coilFractionMin=0.0, coilFractionMax=1.0, exclusive=False):

        self.helixFractionMax = helixFractionMax
        self.helixFractionMin = helixFractionMin
        self.sheetFractionMax = sheetFractionMax
        self.sheetFractionMin = sheetFractionMin
        self.coilFractionMax = coilFractionMax
        self.coilFractionMin = coilFractionMin
        self.exclusive = exclusive

    def __call__(self, t):
        structure = t[1]
        contains_polymer = False
        global_match = False
        num_chains = structure.chains_per_model[0]
        sec_struct = structure.sec_struct_list
        group_counter = 0

        for i in range(num_chains):
            helix = 0.0
            sheet = 0.0
            coil = 0.0
            other = 0.0
            match = True

            chain_type = [chain['type'] for chain in structure.entity_list
                          if i in chain['chainIndexList']][0]
            polymer = chain_type == 'polymer'

            if polymer:
                contains_polymer = True
            else:
                match = False

            for j in range(structure.groups_per_chain[i]):

                if match and polymer:
                    code = sec_struct[group_counter]
                    secondary_structure = DsspSecondaryStructure.get_q3_code(
                        code)

                    if secondary_structure == DsspSecondaryStructure.ALPHA_HELIX:
                        helix += 1
                    elif secondary_structure == DsspSecondaryStructure.EXTENDED:
                        sheet += 1
                    elif secondary_structure == DsspSecondaryStructure.COIL:
                        coil += 1
                    else:
                        other += 1

                group_counter += 1

            if match and polymer:
                n = structure.groups_per_chain[i] - other
                helix /= n
                sheet /= n
                coil /= n

                match = helix >= self.helixFractionMin and \
                    helix <= self.helixFractionMax and \
                    sheet >= self.sheetFractionMin and \
                    sheet <= self.sheetFractionMax and \
                    coil >= self.coilFractionMin and \
                    coil <= self.coilFractionMax

            if (polymer and match and not self.exclusive):
                return True

            if (polymer and not match and self.exclusive):
                return False

            if match:
                global_match = True

        return global_match and contains_polymer



#!/user/bin/env python
'''swissModelDataset

This module provides access to SWISS-MODEL datasets containing homology models.

References
----------
    SWISS-MODEL API:
        https://swissmodel.expasy.org/docs/repository_help#smr_api
    Bienert S, Waterhouse A, de Beer TA, Tauriello G, Studer G, Bordoli L,
    Schwede T (2017). The SWISS-MODEL Repository - new features and
    functionality, Nucleic Acids Res. 45(D1):D313-D319.
        https://dx.doi.org/10.1093/nar/gkw1132
    Biasini M, Bienert S, Waterhouse A, Arnold K, Studer G, Schmidt T, Kiefer F,
    Gallo Cassarino T, Bertoni M, Bordoli L, Schwede T(2014). The SWISS-MODEL
    Repository - modelling protein tertiary and quaternary structure using
    evolutionary information, Nucleic Acids Res. 42(W1):W252–W258.
        https://doi.org/10.1093/nar/gku340

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import requests
import tempfile
from pyspark.sql import SparkSession
from pyspark.sql.functions import col, explode

SWISS_MODEL_REST_URL = "https://swissmodel.expasy.org/repository/uniprot/"
SWISS_MODEL_PROVIDER = ".json?provider=swissmodel"
PDB_PROVIDER = ".json?provider=pdb"


def get_swiss_models(uniProtIds):
    '''Downloads metadata for SWISS-MODEL homology models for alist of
    UniProtIds. The original data schema is flatterened into a row-based schema.

    Example
    -------
        uniProtIds = ["P36575", "P24539", "O00244"]
        ds = swissProtDataset.get_swiss_models(uniProtIds)
        ds.show()

        +------+--------+----+---+-----+----------+----+--------+-----------+--------+--------+--------+----------+-----------+
        |    ac|sequence|from| to|qmean|qmean_norm|gmqe|coverage|oligo-state|  method|template|identity|similarity|coordinates|
        +------+--------+----+---+-----+----------+----+--------+-----------+--------+--------+-- -----+----------+-----------+
        |P36575|MSKVF...|   2|371|-3.06|0.66345522|0.75|0.953608|    monomer|Homology|1suj.1.A|68.66484|0.50463312|https://...|
        |P24539|MLSRV...|  76|249|-2.51|0.67113881|0.65|0.679687|    monomer|Homology|5ara.1.S|84.48275|0.54788881|https://...|
        |O00244|MPKHE...|   1| 68| 1.04|0.84233218|0.98|     1.0| homo-2-mer|Homology|1fe4.1.A|   100.0|0.60686457|https://...|
        +------+--------+----+---+-----+----------+----+--------+-----------+--------+--------+--------+----------+-----------+

    Attributes
    ----------
        uniProtIds (list): list of UniProt Ids

    Returns
    -------
        SwissModel dataset
    '''

    dataset = get_swiss_models_raw_data(uniProtIds)
    return _flatten_dataset(dataset)


def get_swiss_models_raw_data(uniProtIds):
    '''Downloads the raw metadata for SWISS-MODEL homology models. This dataset
    is in the original data schema as downloaded from SWISS-MODEL.

    Attributes
    ----------
        uniProtIds (list): list of UniProt Ids

    Returns
    -------
        SwissModel dataset in original data schema
    '''

    paths = []
    for uniProtId in uniProtIds:
        url = SWISS_MODEL_REST_URL + uniProtId + SWISS_MODEL_PROVIDER
        req = requests.get(url)

        inputStream = req.content
        # TODO temporary solution for inability to retrive data
        if (len(inputStream) - len(uniProtId)) < 103:
            print(f"WARNING: Counld not load data for: {uniProtId}")
            continue

        # save data to temporary file requires as input to dataset reader
        paths.append(_save_temp_file(inputStream.decode("utf-8")))

    # load temporary JSON files into Spark dataset
    dataset = _read_json_files(paths)
    return dataset


def _flatten_dataset(ds):
    '''Flattens the original hierarchical data schema into a simple row-based
    schema. Some less useful data are excluded.

    Attributes
    ----------
        ds (dataset): the original spark dataset

    Returns
    -------
        flattened dataset
    '''

    ds = ds.withColumn("structures", explode(ds.result.structures))
    return ds.select(col("query.ac"), col("result.sequence"), \
                     col("structures.from"), col("structures.to"), \
                     col("structures.qmean"), col("structures.qmean_norm"), \
                     col("structures.gmqe"), col("structures.coverage"), \
                     col("structures.oligo-state"), col("structures.method"), \
                     col("structures.template"), col("structures.identity"), \
                     col("structures.similarity"), col("structures.coordinates"),\
                     col("result.md5"), col("structures.md5"))

def _save_temp_file(inputStream):
    '''Saves tabular report as a temporary CSV file

    Attributes
    ----------
        inputStream (str): inputStream from swiss model

    Returns
    -------
        path to the tempfile
    '''
    tempFile = tempfile.NamedTemporaryFile(delete=False)
    with open (tempFile.name, "w") as t:
        t.writelines(inputStream)
    return tempFile.name


def _read_json_files(paths):
    '''Reads a list of json files to Spark dataset

    Attributes
    ----------
        paths (list): a list of paths to temporary json files

    Returns
    -------
        a sparkdataset
    '''
    spark = SparkSession.builder.getOrCreate()
    dataset = spark.read \
                   .format("json") \
                   .load(paths)
    return dataset



#!/user/bin/env python
'''rFree.py:

This filter returns true if the rFree value for this structure is within the
specified range

References
----------
    rFree:
        http://pdb101.rcsb.org/learn/guide-to-understanding-pdb-data/r-value-and-r-free

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''


class RFree(object):
    '''This filter returns True if the rFree value for this structure in withing
    the specified range.

    Attributes
    ----------
        min_Rfree (float): The lower bound r_free value
        max_RFree (float): The upper bound r_free value
    '''

    def __init__(self, minRfree, maxRfree):
        self.min_Rfree = minRfree
        self.max_Rfree = maxRfree

    def __call__(self, t):
        if t[1].r_free == None:
            return False

        return (t[1].r_free >= self.min_Rfree and t[1].r_free <= self.max_Rfree)



#!/user/bin/env python
'''containsSequenceRegex.py:

This filter returns true if the polymer sequence motif matches the specified regular expression.
Sequence motifs support the following one-letter codes:
20 standard amino acids,
O for Pyrrolysine,
U for Selenocysteine,
X for non-standard amino acid

Reference
---------
    Sequence motif:
        https://en.wikipedia.org/wiki/Sequence_motif

Examples
--------
    Short sequence fragmen -- NPPTP:
        The motif search supports wildcard queries by placing a '.' at the
        variable residue position. A query for an SH3 domains using the
        consequence sequence -X-P-P-X-P (where X is a variable residue and P is
        Proline),can be expressed as: .PP.P

    Ranges of variable residues are specified by the {n} notation, where n is
    the number of variable residues. To query a motif with seven variables
    between residues W and G and twenty variable residues between G and L use
    the following notation:
        W.{7}G.{20}L

    Variable ranges are expressed by the {n,m} notation, where n is the minimum
    and m the maximum number of repetitions. For example the zinc finger motif
    that binds Zn in a DNA-binding domain can be expressed as:
        C.{2,4}C.{12}H.{3,5}H

    The '^' operator searches for sequence motifs at the beginning of a protein
    sequence. The following two queries find sequences with N-terminal Histidine
    tags:
        ^HHHHHH or ^H{6}

    Square brackets specify alternative residues at a particular position.
    The Walker (P loop) motif that binds ATP or GTP can be expressed as:
        [AG].{4}GK[ST]
        A or G are followed by 4 variable residues, then G and K, and finally
        S or T

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import re


class ContainsSequenceRegex(object):
    '''This filter returns true if the polymer sequence motif matches the
    specified regular expression.

    Attributes
    ----------
        regularExpression (String): The regular expression of protein sequence
    '''

    def __init__(self, regularExpression):
        self.regex = regularExpression

    def __call__(self, t):
        structure = t[1]
        entity_list = [b['sequence'] for b in structure.entity_list]

        # This filter passes only single chains and the sequence cannot be empty
        for entity in entity_list:
            if len(entity) > 0:
                if len(re.findall(self.regex, entity)) > 0:
                    return True
        return False



from .rWork import RWork
from .rFree import RFree
from .resolution import Resolution
from .experimentalMethods import ExperimentalMethods
from .containsSequenceRegex import ContainsSequenceRegex
from .containsGroup import ContainsGroup
from .containsPolymerChainType import ContainsPolymerChainType
from .containsAlternativeLocations import ContainsAlternativeLocations
from .containsDProteinChain import ContainsDProteinChain
from .containsLProteinChain import ContainsLProteinChain
from .containsRnaChain import ContainsRnaChain
from .containsDnaChain import ContainsDnaChain
from .containsDSaccharideChain import ContainsDSaccharideChain
from .secondaryStructure import SecondaryStructure
from .depositionDate import DepositionDate
from .releaseDate import ReleaseDate
from .polymerComposition import PolymerComposition
from .orFilter import OrFilter
from .notFilter import NotFilter



#!/user/bin/env python
'''coordinateGeometry.py

This class calculates distances, angles, and orientational order parameters
for atoms coordinated to a central atom

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

import numpy as np
import itertools as it
import math


class CoordinateGeometry(object):
    '''A coordinate geoetry class that calcualtes distances angles and
    orientational order parameters.

    Attributes
    ----------
        center : coordinates of the center atom
        neighbors (list): coordinates of neighbor atoms
    '''

    def __init__(self, center, neighbors):

        self.center = center
        self.neighbors = neighbors
        self._calc_distances()
        self._calc_angles()
        self._calc_dot_products()

    def get_distance(self):
        '''Returns an array of distances from the coordination center to the
        neighbor atoms.

        Returns
        -------
            Numpy array of pairwise angles
        '''

        return self.distances

    def get_angles(self):
        '''Returns all pairwise angles between a center and pairs of neighbor atoms.

        Returns
        -------
            Numpy array of pairwise angles
        '''

        return self.angles

    def q3(self):
        '''Returns a normalized trigonal orientational order parameter q3. The
        orientational order parameter q3 measures the extent to which a molecule
        and its three nearest neighbors adopt a trigonal arrangement. It is equal
        to 0 for a random arrangement and equals 1 in a perfectly trigonal
        arrangement.

        Reference
        ---------
            Richard H. Henchman and Stuart J. Cockramb (2013), Water’s
            Non-Tetrahedral Side, Faraday Discuss., 167, 529.
            <a href="https://dx.doi.org/10.1039/c3fd00080j">doi:10.1039/c3fd00080j</a>

        Returns
        -------
            trigonal orientational order parameter
        '''

        if len(self.neighbors) < 3:
            raise ValueError(f"ERROR: Trigonality calculation requires at least \
                             3 neighbors, but found: {len(self.neighbors)}")

        total = 0.0
        for i, j in it.combinations(range(3), 2):
            total += (self.dotProducts[i][j] + 0.5) ** 2

        return float(1.0 - 4.0 / 7.0 * total)

    def q4(self):
        '''Returns a normalized tetrahedral orientational order parameter q4. The
        orientational order parameter q4 measures the extent to which a central
        atom and its four nearest neighbors adopt a tetrahedral arrangement. It
        is equal to 0 for a random arrangement and equals 1 in a perfectly
        tetrahedral arrangement. It can reach a minimum value of -3 for unusual
        arrangements.

        Reference
        ---------
            Jeffrey R. Errington & Pablo G. Debenedetti (2001) Relationship
            between structural order and the anomalies of liquid water,
            Nature 409, 318-321.
            <a href="https://dx.doi.org/10.1038/35053024">doi:10.1038/35053024</a>

            P.-L. Chau & A. J. Hardwick (1998) A new order parameter for
            tetrahedral configurations, Molecular Physics, 93:3, 511-518.
            <a href"https://dx.doi.org/10.1080/002689798169195">doi:10.1080/002689798169195</a>

        Returns
        -------
            tetrahedra orientational order parameter
        '''

        if len(self.neighbors) < 4:
            raise ValueError(f"ERROR: Tetrahedral calculation requires at least \
                             4 neighbors, but found: {len(self.neighbors)}")

        total = 0.0
        for i, j in it.combinations(range(4), 2):
            total += (self.dotProducts[i][j] + 1.0 / 3.0) ** 2

        return float(1.0 - 3.0 / 8.0 * total)

    def q5(self):
        '''Returns a normalized trigonal bipyramidal orientational order parameter
        q5. The first three nearest atoms to the center atom are used to define
        the equatorial positions. The next two nearest atoms are used to specify
        the axial positions. The orientational order parameter q5 measures the
        extent to which the five nearest atoms adopt a trigonal bipyramidal
        arrangement. It is equal to 0 for a random arrangement and equals 1 in a
        perfectly trigonal bipyramidal arrangement. It can reach negative values
        for certain arrangements.

        Reference
        ---------
            Richard H. Henchman and Stuart J. Cockramb (2013), Water’s
            Non-Tetrahedral Side, Faraday Discuss., 167, 529. <a href=
            "https://dx.doi.org/10.1039/c3fd00080j">doi:10.1039/c3fd00080j</a>
            Note, the summations in equation (3) in this paper is incorrect.
            This method uses the corrected version (R. Henchman, personal
            communication).

        Returns
        -------
            trigonal bipyramidal orientational order parameter
        '''

        if len(self.neighbors) < 5:
            raise ValueError(f"ERROR: Tetrahedral calculation requires at least \
                             5 neighbors, but found: {len(self.neighbors)}")

        # 3 equatorial-equatorial angles (120 deg: cos(120) = -1/2)
        ee = [(0, 1), (0, 2), (1, 2)]
        sum1 = sum([(self.dotProducts[i][j] + 1.0 / 2.0) ** 2 for i, j in ee])

        # 6 equatorial-axial angles (90 deg: cos(90) = 0)
        ea = [(0, 3), (0, 4), (1, 3), (1, 4), (2, 3), (2, 4)]
        sum2 = sum([self.dotProducts[i][j] ** 2 for i, j in ea])

        # 1 axial-axial angle (180 deg: cos(180) = -1)
        sum3 = (self.dotProducts[3][4] + 1) ** 2

        return float(1.0 - 6.0 / 35.0 * sum1 - 3.0 / 10.0 * sum2 - 3.0 / 40.0 * sum3)

    def q6(self):
        '''Returns a normalized octahedra orientational order parameter q6.
        The orientational order parameter q6 measures the extent to which a
        central atom and its six nearest neighbors adopt an octahedral arrangement.
        It is equal to 0 for a random arrangement and equals 1 in a perfectly
        octahedralhedral arrangement. It can reach negative values for certain
        arrangements.

        Reference
        ---------
            Richard H. Henchman and Stuart J. Cockramb (2013), Water’s
            Non-Tetrahedral Side, Faraday Discuss., 167, 529.
            <a href= "https://dx.doi.org/10.1039/c3fd00080j">doi:10.1039/c3fd00080j</a>
            The same method as described in this paper was used to derive the q6
            parameter (R. Henchman, personal communication).

        Returns
        -------
            Octahedral orientational order parameter
        '''

        if len(self.neighbors) < 6:
            raise ValueError(f"ERROR: Octrahedrality calculation requires at \
                             least 5 neighbors, but found: {len(self.neighbors)}")

        # 4 angles between positions in equatorial plane
        equatorial = [(0, 1), (1, 2), (2, 3), (3, 0)]
        sum1 = sum([self.dotProducts[i][j]**2 for i, j in equatorial])

        # 4 angles between axial position 1 and equatorial plane
        axial_1 = [(0, 4), (1, 4), (2, 4), (3, 4)]
        sum2 = sum([self.dotProducts[i][j]**2 for i, j in axial_1])

        # 4 angles between axial position 2 and equatorial plane
        axial_2 = [(0, 5), (1, 5), (2, 5), (3, 5)]
        sum3 = sum([self.dotProducts[i][j]**2 for i, j in axial_2])

        return float(1.0 - 1.0 / 4.0 * sum([sum1, sum2, sum3]))

    def _calc_distances(self):
        '''Calcualte and set the distances between the center and it's neighbors
        '''

        dist = [float(np.linalg.norm(self.center - n)) for n in self.neighbors]
        self.distances = dist

    def _calc_angles(self):
        '''Calcualte and set the angles between the center and it's neighbors
        '''

        self._get_vectors()
        self.angles = [0.0] * \
            int((len(self.neighbors) * len(self.neighbors) - 1) / 2)

        n = 0
        for i in range(len(self.vectors) - 1):
            for j in range(i + 1, len(self.vectors)):
                self.angles[n] = self._angle(self.vectors[i], self.vectors[j])
                n += 1

    def _calc_dot_products(self):
        '''Calcualte and set the dot products between the center and it's neighbors
        '''

        index = self._get_index_by_distance(self.distances)
        vectors = np.array([(self.center - self.neighbors[index[i]]) /
                            np.linalg.norm(
                                self.center - self.neighbors[index[i]])
                            for i in range(len(self.neighbors))])

        self.dotProducts = np.zeros((len(vectors), len(vectors)))
        for i in range(len(vectors) - 1):
            for j in range(i + 1, len(vectors)):
                self.dotProducts[i, j] = float(np.dot(vectors[i], vectors[j]))

    def _get_vectors(self):
        '''Set the center to neighbors vectors
        '''

        self.vectors = [self.center - n for n in self.neighbors]

    def _angle(self, a, b):
        '''Calculate the angle between two points

        Attributes
        ----------
            a: point a
            b: point b
        '''
        arccosInput = np.dot(a, b) / np.linalg.norm(a) / np.linalg.norm(b)
        arccosInput = 1.0 if arccosInput > 1.0 else arccosInput
        arccosInput = -1.0 if arccosInput < -1.0 else arccosInput

        return float(math.acos(arccosInput))

    def _get_index_by_distance(self, values):
        '''Get a list of indices sorted by their calculated distance
        '''

        indexed_values = [(index, value) for index, value in enumerate(values)]
        indexed_values.sort(key=lambda x: x[1])
        return [tup[0] for tup in indexed_values]



#!/user/bin/env python
'''blastCluster.py

This filter passes through representative structures from the RCSB PDB
BlastCLust cluster. A sequence identity thresholds needs to be specified.
The representative for each cluster is the first chain in a cluster.

References
----------
    BlastClust cluster field names:
        http://www.rcsb.org/pdb/statistics/clusterStatistics.do

Example
-------
    Find representative PDB entries at 90% sequence identity:
        sequenceIdentity = 90
        pdb = pdb.filter(BlastCluster(90))

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

import urllib.request

class BlastCluster(object):
    '''Filters blast clusters

    Attributes
    ----------
        sequenceIdentity (int): sequence indentity for blast
    '''
    def __init__(self, sequenceIdentity):

        clusters = self.get_blast_cluster(sequenceIdentity)

        self.pdbIds = set()

        for protein in clusters:
            self.pdbIds.add(protein)
            self.pdbIds.add(protein[:4])


    def __call__(self, t):
        return t[0] in self.pdbIds


    def get_blast_cluster(self, sequenceIdentity):

        if sequenceIdentity not in [30,40,50,70,90,95,100]:
            raise Exception(f"Error: representative chains are not availible for \
                            sequence Identity {sequenceIdentity}.\n Must be in \
                            range [30,40,50,70,90,95,100]")
            return

        coreUrl = "ftp://resources.rcsb.org/sequence/clusters/"
        clusters = []
        inputStream = urllib.request.urlopen(f"{coreUrl}bc-{sequenceIdentity}.out")

        for line in inputStream:
            line = str(line)[2:-3].replace("_",".").strip("\\n")
            clusters += line.split(" ")

        return clusters



#!/user/bin/env python
'''atomInteraction.py:

AtomInteraction contains interaction information of a central atom with it's
interacting neighbors (coordination sphere). Once this data structure is filled
with interaction centers, this class calculates various geometric properties
such as distance, angles, and order parameters for the interacting atoms.
Finally, it provides methods for creating row-rise representations of the data
in Spark Datasets.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''

from mmtfPyspark.interactions import CoordinateGeometry, InteractionCenter
from pyspark.sql import Row
from pyspark.sql.types import *
import numpy as np


class AtomInteraction(object):

    q3 = None
    q4 = None
    q5 = None
    q6 = None
    distances = None
    angles = None

    def __init__(self):

        self.structure_id = None
        self.center = None
        self.neighbors = []

    def set_structure_id(self, structureId):
        '''Sets the structure identifier

        Attributes
        ----------
            structureId (str): the structure identifier
        '''

        self.structureId = structureId

    def get_structure_id(self):
        '''Returns the structure identifier

        Returns
        -------
            structure identifier
        '''

        return self.structureId

    def set_center(self, center):
        '''Sets the central atom information of a coordination sphere.

        Attributes
        ----------
            center (InteractionCenter): central atom information
        '''

        self.center = center

    def get_center(self):
        '''Returns information about the central atom of a coordination sphere

        Returns
        -------
            centeral atom information
        '''

        return self.center

    def add_neighbor(self, neighbor):
        '''Adds a neighbor interaction center.

        Attributes
        ----------
            neighbor (InteractionCenter): an interation with the central atom
        '''

        self.neighbors.append(neighbor)

    def get_interactions(self):
        '''Returns information about the interacting neighbor atoms.

        Returns
        -------
            interaction centers
        '''

        return self.neighbors

    def get_num_interactions(self):
        '''Returns the number of neighboring atoms that interact with the central
        atom

        Returns
        -------
            number of neighboring atoms that interact with the central atom
        '''

        return len(self.neighbors)

    def calc_coordination_geometry(self, maxInteraction):
        '''Calculates geometric properties of the coordination sphere.
        The geometric properties include orientational order parameters that
        describe the arrangement of the atoms in the coordination sphere,
        distances and angles of the neighcor atoms with the cnter atom.

        Attributes
        ----------
            maxInteraction (int): maximum number of interaction
        '''

        neighborPoints = [n.get_coordinates() for n in self.neighbors
                          if n.get_coordinates() is not None]

        geom = CoordinateGeometry(
            self.center.get_coordinates(), neighborPoints)

        # calculate distances to the central atom
        #self.distances = np.empty(maxInteraction)
        self.distances = [0.0] * maxInteraction
        for i, dist in enumerate(geom.get_distance()):
            self.distances[i] = dist

        # calculate angles among all interacting atoms with the central atom
        numInteraction = int(maxInteraction * (maxInteraction - 1) / 2)
        ang = geom.get_angles()
        self.angles = [np.NaN] * numInteraction
        self.angles[:len(ang[:numInteraction])] = ang[:numInteraction]

        # TODO: points or neighbor points
        if len(neighborPoints) > 2:
            self.q3 = geom.q3()
        if len(neighborPoints) > 3:
            self.q4 = geom.q4()
        if len(neighborPoints) > 4:
            self.q5 = geom.q5()
        if len(neighborPoints) > 5:
            self.q6 = geom.q6()

    def get_multiple_interactions_as_row(self, maxInteractions):
        '''Returns interactions and geometric information in a single row

        Returns
        -------
            row of itneractions and geometric information
        '''
        while self.get_num_interactions() < maxInteractions:
            self.neighbors.append(InteractionCenter())

        self.length = InteractionCenter.get_length()

        self.calc_coordination_geometry(maxInteractions)

        data = [self.structureId, self._get_number_of_polymer_chains(),
                self.q3, self.q4, self.q5, self.q6]

        # Copy data for query atom
        data += self.center.get_as_object()

        # Copy data for interaction atoms
        for i, neighbor in enumerate(self.neighbors):
            data += neighbor.get_as_object()
            data.append(self.distances[i])

        data += self.angles

        return Row(data)

    def get_pair_interactions_as_rows(self):
        '''Return rows of pairwise interactions with the central atom

        Returns
        -------
            rows of pairwise interactions with the central atom
        '''

        rows = []
        length = InteractionCenter.get_length()
        self.calc_coordination_geometry()

        for i, neighbor in enumerate(self.neighbors):
            index = 0

        return rows

    def get_schema(self, maxInteractions):
        '''Returns the schema for a row of atom interaction inforamtion. The schema
        is used to create a Dataset<Row> from the row information.

        Returns
        -------
            return schema for dataset
        '''

        sf = []
        sf.append(StructField("pdbId", StringType(), False))
        sf.append(StructField("polyChains", IntegerType(), False))
        sf.append(StructField("q3", FloatType(), True))
        sf.append(StructField("q4", FloatType(), True))
        sf.append(StructField("q5", FloatType(), True))
        sf.append(StructField("q6", FloatType(), True))

        # Copy schema for query atom
        sf += InteractionCenter.get_struct_fields(0)

        # Copy schema info for interacting atoms and their distances
        for i in range(maxInteractions):
            sf += InteractionCenter.get_struct_fields(i + 1)
            sf.append(StructField(f"distance{i+1}", FloatType(), True))

        # Add schema for angles
        for i in range(maxInteractions - 1):
            for j in range(i + 1, maxInteractions):
                sf.append(StructField(f"angle{i+1}-{j+1}", FloatType(), True))

        return StructType(sf)

    def get_pair_interaction_schema(self):
        '''Returns the schema for a row of pairwise atom interactions.
        The schema is used to create a Dataset<Row> from the row information

        Return
        ------
            return schema for dataset
        '''

        sf = []
        sf.append(StructField("pdbId", StringType(), False))

        # copy schema info for query atom
        sf += InteractionCenter.get_struct_fields(0)

        # copy schema infor for interacting atoms and their distnce
        sf += InteractionCenter.get_struct_fields(1)
        sf.append(StructField("distance1", FloatType(), True))

        return StructType(sf)

    def _get_number_of_polymer_chains(self):
        '''Returns the number of unique polymer chains in the coordination sphere.

        Returns
        -------
            number of unique polymer chains in the coordination sphere.
        '''

        return len({center.get_chain_name for center in self.neighbors
                    if center.atomName is not None and center.get_sequence_position() >= 0})

    def _get_num_columns(self, maxInteractions):
        '''Returns the number of columns in a Row

        The number of columns:
            structureId + polymerChains + q3 + q4 + q5 + q6: 6
            query + interaction centers: (maxInteractions + 1) * length
            distances: maxInteractions
            angles: maxInteractions * (maxInteractions - 1) / 2

        Attributes
        ----------
            maxInteractions (int): maximum number of interactions

        Returns
        -------
            number of columns in a row
        '''

        numAngles = maxInteractions * (maxInteractions - 1) / 2
        return int(6 + (maxInteractions + 1) * self.length + maxInteractions * numAngles)

    def _set(self, data, value, i):
        '''Sets the ith index of data to value and increase i by one
        '''

        data[i] = value
        i += 1
        return data, i



#!/user/bin/env python
'''containsAlternativeLocations.py

This filter return true if this structure contains an alternative location

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''


class ContainsAlternativeLocations(object):

    def __call__(self, t):
        structure = t[1]

        if not structure.alt_loc_set:
            structure = structure.set_alt_loc_list()    

        for c in structure.alt_loc_list:
            if c != '\0':
                return True

        return False



#!/usr/bin/env python
'''sparkMultiClassClassifier.py

Fits a multi-class classification model using mllib classification method and
returns classification metrics.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from pyspark.ml.feature import StringIndexer, IndexToString
from pyspark.ml import Pipeline
from pyspark.mllib.evaluation import BinaryClassificationMetrics, MulticlassMetrics
from collections import OrderedDict
import time


class SparkMultiClassClassifier(object):
    '''Fits a multi-class classification model using mllib classification method and
    returns classification metrics.

    Attributes
    ----------
        predictor: type of multi-class classifier
        label (str): classification label
        testFraction (float): test set fraction [0.3]
        seed (int): random seed
    '''

    def __init__(self, predictor, label, testFraction=0.3, seed=1):

        self.predictor = predictor
        self.label = label
        self.testFraction = testFraction
        self.seed = seed

    def fit(self, data):
        '''Dataset must at least contain the following two columns:
        label: the class labels
        features: feature vector

        Attributes
        ----------
            data (Dataset<Row>): input data

        Returns
        -------
            map with metrics
        '''

        start = time.time()

        classCount = int(data.select(self.label).distinct().count())

        labelIndexer = StringIndexer().setInputCol(self.label) \
                                      .setOutputCol("indexedLabel") \
                                      .fit(data)

        # Split the data into training and test sets (30% held out for testing)
        splits = data.randomSplit(
            [1.0 - self.testFraction, self.testFraction], self.seed)
        trainingData = splits[0]
        testData = splits[1]

        labels = labelIndexer.labels


        print("\n Class\tTrain\tTest")
        for l in labels:
            print("%s\t%i\t%i" % (l \
                                  ,(trainingData.filter(trainingData[self.label] == l)).count() \
                                  ,(testData.filter(testData[self.label] == l)).count() \
                                  )
                  )

        # Set input columns
        self.predictor.setLabelCol("indexedLabel").setFeaturesCol("features")

        # Convert indexed labels back to original labels
        labelConverter = IndexToString().setInputCol("prediction") \
                                        .setOutputCol("predictedLabel") \
                                        .setLabels(labelIndexer.labels)

        # Chain indexers and forest ina Pipline
        pipeline = Pipeline().setStages([labelIndexer, self.predictor, labelConverter])

        # Train model. This also runs the indexers
        model = pipeline.fit(trainingData)

        # Make predictions
        predictions = model.transform(testData).cache()

        # Display some sample predictions
        print(f"\nSample predictions: {str(self.predictor).split('_')[0]}") # TODO predictor.getClass().getSimpleName()
        predictions.sample(False, 0.1, self.seed).show(5)

        predictions = predictions.withColumnRenamed(self.label, "stringLabel")
        predictions = predictions.withColumnRenamed("indexedLabel", self.label)

        # Collect metrics

        pred = predictions.select("prediction", self.label)

        metrics = OrderedDict()
        metrics["Method"] = str(self.predictor).split('_')[0]

        if classCount == 2:
            b = BinaryClassificationMetrics(pred.rdd)
            metrics["AUC"] = str(b.areaUnderROC)
        m = MulticlassMetrics(pred.rdd)
        metrics["F"] = str(m.weightedFMeasure())
        metrics["Accuracy"] = str(m.accuracy)
        metrics["Precision"] = str(m.weightedPrecision)
        metrics["Recall"] = str(m.weightedRecall)
        metrics["False Positive Rate"] = str(m.weightedFalsePositiveRate)
        metrics["True Positive Rate"] = str(m.weightedTruePositiveRate)
        metrics[""] = f"\nConfusion Matrix\n{labels}\n{m.confusionMatrix()}"

        end = time.time()
        print(f"Total time taken: {end-start}\n")

        return metrics



#!/user/bin/env python
'''polymerComposition.py

This filter returns entries that contain chains made of the specified
monomer types.

authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''


class PolymerComposition(object):
    '''The default constructor returns entries that contain at least
    one chain that matches the conditions. If the "exclusive" flag is set to true
    in the constructor, all chains must match the conditions. For a multi-model
    structure, this filter only checks the first model.

    Optional constructor that can be used to filter entries that exclusively
    match all chains.

    Attributes
    ----------
        exclusive (bool): if true, all chains must be made of the specified
                          monomers
    '''

    # define sets of residue types eg:
    # 20 nat. amino acids, 22 nat. amino acids, nat. DnNA, nat. RNA
    # custom sets
    AMINO_ACIDS_20 = ["ALA", "ARG", "ASN", "ASP", "CYS", "GLN", "GLU", "GLY", "HIS",
                      "ILE", "LEU", "LYS", "MET", "PHE", "PRO", "SER", "THR", "TRP", "TYR", "VAL"]
    AMINO_ACIDS_22 = ["ALA", "ARG", "ASN", "ASP", "CYS", "GLN", "GLU", "GLY", "HIS", "ILE",
                      "LEU", "LYS", "MET", "PHE", "PRO", "SER", "THR", "TRP", "TYR", "VAL", "SEC", "PYL"]
    DNA_STD_NUCLEOTIDES = ["DA", "DC", "DG", "DT"]
    RNA_STD_NUCLEOTIDES = ["A", "C", "G", "U"]

    def __init__(self, monomer_type, exclusive=False):
        if type(monomer_type) == str:
            monomer_type = monomer_type.split(",")

        self.exclusive = exclusive
        self.residues = monomer_type

    def __call__(self, t):
        structure = t[1]
        contains_polymer = False
        global_match = False
        # get number of chains in first model, nessary?
        num_chains = structure.chains_per_model[0]
        # chain types in entity as key, enetity from entity_list
        group_counter = 0

        for i in range(num_chains):
            match = True
            chain_type = [chain['type'] for chain in structure.entity_list
                          if i in chain['chainIndexList']][0]
            polymer = chain_type == "polymer"

            if polymer:
                contains_polymer = True
            else:
                match = False
            # group_type_list

            for j in range(structure.groups_per_chain[i]):
                if match and polymer:
                    group_idx = structure.group_type_list[group_counter]
                    group_type = structure.group_list[group_idx]['groupName']
                    match = (group_type in self.residues)
                group_counter += 1

            if (polymer and match and not self.exclusive):
                return True

            if (polymer and not match and self.exclusive):
                return False

            if match:
                global_match = True

        return global_match and contains_polymer



#!/user/bin/env python
'''structureToAllInteractions.py:

Finds interactions of a specified group within a specified cutoff distance

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from pyspark.sql import Row


class StructureToAllInteractions(object):
    '''Class that finds interactions of a specified group within a specified
    cutoff distance

    Attributes
    ----------
        groupName (str): specified group in structure
        cutoffDistance (float): cutoff distance used during search
    '''

    def __init__(self, groupName, cutoffDistance):
        self.groupName = groupName
        self.cutoffDistance = cutoffDistance

    def __call__(self, t):
        structureId = t[0]
        structure = t[1]

        groupIndices, groupNames = self._get_group_indices(structure)
        interactions = []

        for i in range(len(groupNames)):

            if groupNames[i] == self.groupName:
                matches = []
                boundingBox = self._calc_bonding_box(structure, groupIndices, i,
                                                     self.cutoffDistance)
                matches += self._find_neighbors(structure,
                                                i, boundingBox, groupIndices)
                interactions += self._get_distance_profile(
                    structureId, matches, i, groupIndices, groupNames, structure)

        return interactions

    def _get_distance_profile(self, structureId, matches, index, groupIndices, groupNames, structure):
        cutoffDistanceSq = self.cutoffDistance * self.cutoffDistance

        x = structure.x_coord_list
        y = structure.y_coord_list
        z = structure.z_coord_list

        first = groupIndices[index]
        last = groupIndices[index + 1]

        groupIndex1 = structure.group_type_list[index]

        rows = []

        for i in matches:
            if i == index:
                continue

            for j in range(groupIndices[i], groupIndices[i + 1]):

                for k in range(first, last):
                    dx = x[j] - x[k]
                    dy = y[j] - y[k]
                    dz = z[j] - z[k]
                    dSq = dx * dx + dy * dy + dz * dz

                    if (dSq <= cutoffDistanceSq):
                        aIndex1 = k - first
                        atomName1 = structure.group_list[groupIndex1]['atomNameList'][aIndex1]
                        element1 = structure.group_list[groupIndex1]['elementList'][aIndex1]

                        groupIndex2 = structure.group_type_list[i]
                        aIndex2 = j - groupIndices[i]
                        atomName2 = structure.group_list[groupIndex2]['atomNameList'][aIndex2]
                        element2 = structure.group_list[groupIndex2]['elementList'][aIndex2]

                        d = dSq ** 0.5
                        row = Row(structureId,
                                  groupNames[index],
                                  atomName1,
                                  element1,
                                  index,
                                  groupNames[i],
                                  atomName2,
                                  element2,
                                  i,
                                  float(d))
                        rows.append(row)
        return rows

    def _find_neighbors(self, structure, index, boundingBox, groupIndices):
        x = structure.x_coord_list
        y = structure.y_coord_list
        z = structure.z_coord_list

        matches = []
        for i in range(len(groupIndices) - 1):

            for j in range(groupIndices[i], groupIndices[i + 1]):

                if (x[j] >= boundingBox[0] and x[j] <= boundingBox[1] and
                    y[j] >= boundingBox[2] and y[j] <= boundingBox[3] and
                        z[j] >= boundingBox[4] and z[j] <= boundingBox[5]):
                    matches.append(i)
                    break

        return matches

    def _calc_bonding_box(self, structure, groupIndices, i, cutoffDistance):
        x = structure.x_coord_list
        y = structure.y_coord_list
        z = structure.z_coord_list

        xMin = -float('inf')
        xMax = float('inf')
        yMin = -float('inf')
        yMax = float('inf')
        zMin = -float('inf')
        zMax = float('inf')

        first = groupIndices[i]
        last = groupIndices[i + 1]

        for i in range(first, last):
            xMin = min(xMin, x[i])
            xMax = max(xMax, x[i])
            yMin = min(yMin, y[i])
            yMax = max(yMax, y[i])
            zMin = min(zMin, z[i])
            zMax = max(zMax, z[i])

        boundingBox = [0] * 6
        boundingBox[0] = float(xMin - cutoffDistance)
        boundingBox[1] = float(xMax + cutoffDistance)
        boundingBox[2] = float(yMin - cutoffDistance)
        boundingBox[3] = float(yMax + cutoffDistance)
        boundingBox[4] = float(zMin - cutoffDistance)
        boundingBox[5] = float(zMax + cutoffDistance)

        return boundingBox

    def _get_group_indices(self, structure):
        '''Creates an atom index to the first atom of each group

        Attributes:
            structure
        '''
        groupIndices, groupNames = [0], []  # Start index for first group
        atomCounter, groupCounter = 0, 0

        # Consider only the first model
        numChains = structure.chains_per_model[0]

        # Loop over all chains
        for i in range(numChains):

            # Loop over all groups in chains
            for j in range(structure.groups_per_chain[i]):
                groupIndex = structure.group_type_list[groupCounter]
                groupNames.append(
                    structure.group_list[groupIndex]['groupName'])
                atomCounter += len(
                    structure.group_list[groupIndex]['atomNameList'])
                groupIndices.append(atomCounter)
                groupCounter += 1

        return groupIndices, groupNames



#!/user/bin/env python
'''structureToSecondaryStructureElements.py:

Maps chain sequences to its sequence segments.

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "done"
'''
from pyspark.sql import Row


class StructureToSecondaryStructureElements(object):
    '''Constructor sets the segment length.

    Attributes
    ----------
        label: label of the structure
        length (int): segment length [4]
    '''

    def __init__(self, label, length=4):
        self.label = label
        self.length = length

    def __call__(self, t):
        sequence = t[1]
        dsspQ3 = t[6]
        sequences = []

        i = 0
        while i < len(sequence):
            currLength = 0
            currSequence = ""

            for j in range(i, len(sequence)):

                if dsspQ3[j:j + 1] == self.label:
                    currLength += 1
                    currSequence += sequence[j: j + 1]
                else:
                    break

            i += currLength + 1

            if currLength >= self.length:
                sequences.append(Row(currSequence, self.label))

            i += 1

        return sequences



#!/user/bin/env python
'''mmtfWriter.py

Encodes and write MMTF encoded structure data to a Hadoop Sequence File

Authorship information:
    __author__ = "Mars (Shih-Cheng) Huang"
    __maintainer__ = "Mars (Shih-Cheng) Huang"
    __email__ = "marshuang80@gmail.com"
    __version__ = "0.2.0"
    __status__ = "Done"
'''

from mmtf.api.mmtf_writer import MMTFEncoder
from mmtfPyspark.utils import MmtfStructure
import gzip
import msgpack
import os
import base64


def write_sequence_file(path, sc, structure, compressed=True):
    '''Encodes and writes MMTF encoded structure data to a Hadoop Sequnce File

    Attributes
    ----------
        path (str): Path to Hadoop file directory)
        sc (Spark context)
        structure (tuple): structure data to be written
        compress (bool): if true, apply gzip compression
    '''
    # Can't apply first() function on list
    if type(structure.first()[1]) == MmtfStructure:
        structure = structure.map(lambda s: (s[0], s[1].set_alt_loc_list()))

    structure.map(lambda t: (t[0], _to_byte_array(t[1], compressed)))\
             .saveAsHadoopFile(path,
                               "org.apache.hadoop.mapred.SequenceFileOutputFormat",
                               "org.apache.hadoop.io.Text",
                               "org.apache.hadoop.io.BytesWritable")


def write_mmtf_files(path, sc, structure):
    '''Encodes and writes MMTF encoded and gzipped structure data to individual .mmtf.gz files.

    Attributes
    ----------
        path (str): Path to Hadoop file directory
        sc (Spark context)
        structure (tuple): structure data to be written
    '''

    if path[-1] != "/":
        path = path + "/"

    if not os.path.exists(path):
        os.makedirs(path)

    structure = structure.map(lambda s: (s[0], s[1].set_alt_loc_list())) \
                         .map(lambda t: (t[0], _to_byte_array(t[1], False))) \
                         .foreach(lambda t: gzip.open(path + t[0] + '.mmtf.gz', mode='wb').write(t[1]))


def to_mmtf_base64(structure):
    '''Encodes a mmtfStructure to base64 byte array

    Attributes
    ----------
        structure (mmtfStructure): structure to be encoded to base64 byte array

    Returns
    -------
        base64 byte array
    '''

    byteArray = _to_byte_array(structure, compressed=False)
    return base64.b64encode(byteArray).decode()


def _to_byte_array(structure, compressed):
    '''Returns an MMTF-encoded byte array with optional gzip compression

    Returns
    -------
        MMTF encoded and optionally gzipped structure data
    '''

    if type(structure) == MmtfStructure:
        if not structure.alt_loc_set:
            structure = structure.set_alt_loc_list()

    byte_array = bytearray(msgpack.packb(MMTFEncoder.encode_data(structure), use_bin_type = True))

    if compressed:
        return gzip.compress(byte_array)
    else:
        return byte_array
